// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Text;

using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Client.Transaction.BuiltIn {
  internal class AuthenticateTransaction : ImapTransactionBase {
    public AuthenticateTransaction(ImapConnection connection, NetworkCredential credential)
      : base(connection)
    {
      if (credential == null)
        throw new ArgumentNullException("credential");

      this.credential = credential;
    }

    public override void Dispose()
    {
      if (saslMechanism != null) {
        saslMechanism.Dispose();
        saslMechanism = null;
      }

      base.Dispose();
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("authentication mechanism name"))
        return ProcessArgumentNotSetted;

      try {
        saslMechanism = SaslClientMechanism.Create((string)Request.Arguments["authentication mechanism name"]);
        saslMechanism.Credential = credential;

        // The service name specified by this protocol's profile of [SASL] is
        // "imap".
        saslMechanism.ServiceName = "imap";

        if (saslMechanism is NTLMMechanism)
          (saslMechanism as NTLMMechanism).TargetHost = Connection.Host;

        return ProcessAuthenticate;
      }
      catch (SaslMechanismNotSupportedException) {
        return ProcessUnsupported;
      }
    }

    private void ProcessArgumentNotSetted()
    {
      FinishError(ImapTransactionResultCode.RequestError, "arguments 'authentication mechanism name' must be setted");
    }

    private void ProcessUnsupported()
    {
      FinishError(ImapTransactionResultCode.RequestError, "unsupported authentication mechanism");
    }

    // 6.2.2. AUTHENTICATE Command
    //    Arguments:  authentication mechanism name
    //    Responses:  continuation data can be requested
    //    Result:     OK - authenticate completed, now in authenticated state
    //                NO - authenticate failure: unsupported authentication
    //                     mechanism, credentials rejected
    //                BAD - command unknown or arguments invalid,
    //                     authentication exchange cancelled
    private void ProcessAuthenticate()
    {
      SendCommand("AUTHENTICATE", ProcessReceiveResponse, Request.Arguments["authentication mechanism name"]);
    }

    protected override void OnCommandContinuationRequestReceived(ImapCommandContinuationRequest continuationRequest)
    {
      // The authentication protocol exchange consists of a series of
      // server challenges and client responses that are specific to the
      // authentication mechanism.  A server challenge consists of a
      // command continuation request response with the "+" token followed
      // by a BASE64 encoded string.  The client response consists of a
      // single line consisting of a BASE64 encoded string.  If the client
      // wishes to cancel an authentication exchange, it issues a line
      // consisting of a single "*".  If the server receives such a
      // response, it MUST reject the AUTHENTICATE command by sending a
      // tagged BAD response.
      var serverChallenge = Smdn.Formats.TextConvert.FromBase64StringToByteArray(continuationRequest.Text);
      byte[] clientResponse;
      var status = saslMechanism.Exchange(serverChallenge, out clientResponse);

      if (status == SaslExchangeStatus.Failed || clientResponse == null)
        SendContinuation("*");
      else
        SendContinuation(Smdn.Formats.TextConvert.ToBase64String(clientResponse));
    }

    private /*readonly*/ NetworkCredential credential;
    private /*readonly*/ SaslClientMechanism saslMechanism;
  }
}
