// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.Threading;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

using Smdn.Net.Imap4.Client;

namespace Smdn.Net.Imap4.Protocol.Client {
  // Layers:
  //   Session.ImapSession
  //   Transaction.IImapTransaction/Transaction.ImapTransactionRequest/Transaction.ImapTransactionResult
  // * Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public class ImapConnection : ImapConnectionBase {
    /*
     * class members
     */
    public const int DefaultTimeoutMilliseconds = 16000;

    public static X509Certificate2Collection CertCollection {
      get { return certCollection; }
    }

    public static RemoteCertificateValidationCallback CertificateValidationCallback {
      get { return certificateValidationCallback; }
      set { certificateValidationCallback = value; }
    }

    public static LocalCertificateSelectionCallback CertificateSelectionCallback {
      get { return certificateSelectionCallback; }
      set { certificateSelectionCallback = value; }
    }

    public static Stream CreateSslStream(ImapConnectionBase connection, Stream baseStream)
    {
      var sslStream = new SslStream(baseStream, false, certificateValidationCallback, certificateSelectionCallback);
      var conn = connection as ImapConnection;

      sslStream.AuthenticateAsClient(conn.Host, certCollection, SslProtocols.Default, true);

      Trace.Log(conn, "created SSL/TLS stream for '{0}'", conn.Host);
      Trace.Log(conn, "  protocol: {0}", sslStream.SslProtocol);
      Trace.Log(conn, "  cipher: {0} strength {1}", sslStream.CipherAlgorithm, sslStream.CipherStrength);
      Trace.Log(conn, "  hash: {0} strength {1}", sslStream.HashAlgorithm, sslStream.HashStrength);
      Trace.Log(conn, "  key exchange: {0} strength {1}", sslStream.KeyExchangeAlgorithm, sslStream.KeyExchangeStrength);
      Trace.Log(conn, "  authenticated: {0}", sslStream.IsAuthenticated);
      Trace.Log(conn, "  authenticated as server: {0}", sslStream.IsServer);
      Trace.Log(conn, "  mutually authenticated: {0}", sslStream.IsMutuallyAuthenticated);
      Trace.Log(conn, "  signed: {0}", sslStream.IsSigned);
      Trace.Log(conn, "  encrypted: {0}", sslStream.IsEncrypted);

      return sslStream;
    }

    private static readonly X509Certificate2Collection certCollection = new X509Certificate2Collection();
    private static RemoteCertificateValidationCallback certificateValidationCallback = null;
    private static LocalCertificateSelectionCallback certificateSelectionCallback = null;

    /*
     * properties
     */
    public override string Host {
      get
      {
        CheckDisposed();
        CheckConnected();
        return host;
      }
    }

    internal new Socket Socket {
      get { return base.Socket; }
    }

    // for '5.4. Autologout Timer'
    public DateTime LastSentTime {
      get
      {
        CheckDisposed();
        return lastSentTime;
      }
    }

    // IMAP4 non-synchronizing literals
    // http://tools.ietf.org/html/rfc2088
    public bool SynchronizeLiterals
    {
      get
      {
        CheckDisposed();
        return Sender.SynchronizeLiterals;
      }
      set
      {
        CheckDisposed();
        Sender.SynchronizeLiterals = value;
      }
    }

    public ImapConnection(string host)
      : this(host, -1, null)
    {
    }

    public ImapConnection(string host, int port)
      : this(host, port, null)
    {
    }

    public ImapConnection(string host, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
      : this(host, -1, createAuthenticatedStreamCallback)
    {
    }

    public ImapConnection(string host, int port, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      Connect(host, port, createAuthenticatedStreamCallback);
    }

    public void Connect(string host)
    {
      Connect(host, -1, null);
    }

    public void Connect(string host, int port)
    {
      Connect(host, port, null);
    }

    public void Connect(string host, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      Connect(host, -1, createAuthenticatedStreamCallback);
    }

    public void Connect(string host, int port, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      if (port == -1) {
        if (createAuthenticatedStreamCallback == null)
          port = ImapDefaultPort.Imap;
        else
          port = ImapDefaultPort.Imaps;
      }

      if (host == null)
        throw new ArgumentNullException("host");
      if (port < IPEndPoint.MinPort || IPEndPoint.MaxPort < port)
        throw new ArgumentOutOfRangeException("port");

      this.host = host;

      Trace.Info("CID:{0} connecting to {1}", Id, host);

      TcpClient client = null;

      try {
        client = new TcpClient();

        client.ReceiveTimeout = DefaultTimeoutMilliseconds;
        client.SendTimeout    = DefaultTimeoutMilliseconds;

        client.Connect(host, port);

        SetClient(client, createAuthenticatedStreamCallback);
      }
      catch {
        if (client != null)
          client.Close();

        throw;
      }

      Trace.Info("CID:{0} connected to {1} ({2})", Id, host, RemoteEndPoint);
    }

#if TRACE
    protected override void Dispose(bool disposing)
    {
      base.Dispose(disposing);

      if (disposing)
        Trace.Info("CID:{0} disconnected", Id);
    }

    protected override ImapBufferedStream CreateBufferedStream(Stream stream)
    {
      return new ImapBufferedStream(new TracingStream(this, stream));
    }

    private class TracingStream : Smdn.IO.InterruptStream {
      public TracingStream(ImapConnection connection, Stream innerStream)
        : base(innerStream)
      {
        this.connection = connection;
      }

      protected override void OnWritten(byte[] src, int offset, int count)
      {
        Trace.LogSent(connection, src, offset, count);
      }

      protected override int OnRead(byte[] dest, int offset, int count)
      {
        Trace.LogReceived(connection, dest, offset, count);

        return count;
      }

      private ImapConnection connection;
    }
#endif

    internal new void UpgradeStream(ImapUpgradeConnectionStreamCallback upgradeStreamCallback)
    {
      base.UpgradeStream(upgradeStreamCallback);
    }

    internal void SetIsIdling(bool isIdling)
    {
      base.IsIdling = isIdling;
    }

    protected override ImapSender CreateSender(ImapBufferedStream stream)
    {
      return new ImapCommandSender(stream, (Sender == null) ? true : Sender.SynchronizeLiterals);
    }

    protected override ImapReceiver CreateReceiver(ImapBufferedStream stream)
    {
      return new ImapResponseReceiver(stream);
    }

    public ImapCommand CreateCommand(string command, params ImapString[] arguments)
    {
      CheckDisposed();
      CheckConnected();

      var comm = new ImapCommand(command, commandTag.ToString("D4"), arguments);

      if (Interlocked.Increment(ref commandTag) == 0x10000)
        Interlocked.Exchange(ref commandTag, 0);

      return comm;
    }

    public ImapCommand CreateContinuingCommand(params ImapString[] arguments)
    {
      CheckDisposed();
      CheckConnected();

      return new ImapCommand(null, null, arguments);
    }

#region "seinding / receiving"
    public interface ICommandContinuationContext {
    }

    private class SendCommandContext : ICommandContinuationContext {
      public ImapCommand Command;
      public bool Sent;

      public SendCommandContext(ImapCommand command)
      {
        this.Command = command;
        this.Sent = false;
      }
    }

    public ICommandContinuationContext SendCommand(ImapCommand command)
    {
      return SendCommand(new SendCommandContext(command));
    }

    public ICommandContinuationContext SendCommand(ICommandContinuationContext context)
    {
      CheckDisposed();
      CheckConnected();

      try {
        var c = context as SendCommandContext;
        var s = Sender as ImapCommandSender;

        if (c.Sent) {
          s.Send();
        }
        else {
          s.Send(c.Command);
          c.Sent = true;
        }

        lastSentTime = DateTime.Now;

        if (s.CommandContinuing)
          return c;
        else
          return null;
      }
      catch (IOException ex) {
        if (ex.InnerException is SocketException)
          throw new ImapConnectionException((ex.InnerException as SocketException).Message, ex.InnerException);
        else
          throw ex;
      }
    }

    public ImapResponse TryReceiveResponse()
    {
      CheckDisposed();
      CheckConnected();

      // return queued response
      if (0 < responseQueue.Count)
        return responseQueue.Dequeue();

      // receive and parse response
      var r = Receiver as ImapResponseReceiver;

      for (;;) {
        try {
          var response = r.ReceiveResponse();

          if (response == null)
            break;

          responseQueue.Enqueue(response);

          if (!r.ResponseContinuing)
            break;
        }
        catch (IOException ex) {
          var socketException = ex.InnerException as SocketException;

          if (socketException == null)
            // unexpected IO error
            throw ex;

          if (IsIdling && socketException.SocketErrorCode == SocketError.WouldBlock) {
            // expected socket exception
            Trace.Log(this, "idling");
            continue;
          }
          else {
            // unexpected socket exception
            throw new ImapConnectionException(socketException.Message, socketException.InnerException);
          }
        }
      }

      // return received response
      if (0 < responseQueue.Count)
        return responseQueue.Dequeue();

      return null;
    }
#endregion

    private string host = null;
    private DateTime lastSentTime = DateTime.MinValue;
    private int commandTag = 0;
    private /*readonly*/ Queue<ImapResponse> responseQueue = new Queue<ImapResponse>();
  }
}
