// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Net;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.WebClient {
  internal class ImapMailboxWebRequest : ImapWebRequest {
    public ImapMailboxWebRequest(Uri requestUri, ImapSessionManager sessionManager)
      : base(requestUri, sessionManager)
    {
    }

    private bool IsRequestAppend {
      get { return string.Equals(ImapWebRequestMethods.Append, Method, StringComparison.InvariantCultureIgnoreCase); }
    }

    protected override ImapWebResponse InternalGetResponse()
    {
      if (!string.Equals(Method, ImapWebRequestMethods.Append, StringComparison.InvariantCultureIgnoreCase)) {
        var response = UnselectMailbox();

        if (response != null)
          return response;
      }

      if (string.IsNullOrEmpty(Method) ||
          string.Equals(Method, ImapWebRequestMethods.Fetch, StringComparison.InvariantCultureIgnoreCase))
        return GetFetchResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Append, StringComparison.InvariantCultureIgnoreCase))
        return GetAppendResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Create, StringComparison.InvariantCultureIgnoreCase))
        return GetCreateResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Delete, StringComparison.InvariantCultureIgnoreCase))
        return GetDeleteResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Expunge, StringComparison.InvariantCultureIgnoreCase))
        return GetExpungeResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Examine, StringComparison.InvariantCultureIgnoreCase))
        return GetExamineResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Rename, StringComparison.InvariantCultureIgnoreCase))
        return GetRenameResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Subscribe, StringComparison.InvariantCultureIgnoreCase))
        return GetSubscribeResponse();
      else if (string.Equals(Method, ImapWebRequestMethods.Unsubscribe, StringComparison.InvariantCultureIgnoreCase))
        return GetUnsubscribeResponse();
      else
        throw new NotSupportedException(string.Format("unsupported request method: {0}", Method));
    }

    private ImapWebResponse GetFetchResponse()
    {
      var response = SelectRequestMailbox();

      if (response != null)
        return response;

      ImapMessage[] messages;

      var result = Session.Fetch(ImapSequenceSet.CreateUidFromSet(1), ImapFetchDataItem.All, out messages);

      if (result.Failed)
        return new ImapWebResponse(result);

      response = new ImapWebResponse(result);

      response.Messages = messages;

      return response;
    }

#region "APPEND"
    public override Stream GetRequestStream()
    {
      return EndGetRequestStream(BeginGetRequestStream(null, null));
    }

    public override IAsyncResult BeginGetRequestStream(AsyncCallback callback, object state)
    {
      if (!IsRequestAppend)
        // throws NotSupportedException
        return base.BeginGetRequestStream(callback, state);

      if (appendMessageBodyStream != null)
        throw new InvalidOperationException("already called");

      appendMessageBodyStream = new ImapAppendMessageBodyStream();

      // Content-Length
      if (0 < ContentLength)
        appendMessageBodyStream.SetLength(ContentLength);

      var session = GetSession();

      // TODO: TRYCREATE when AllowCreateMailbox == true
      appendAsyncResult = session.BeginAppend(appendMessageBodyStream, DateTimeOffset.Now, null, RequestMailbox);

      return appendAsyncResult;
    }

    public override Stream EndGetRequestStream(IAsyncResult asyncResult)
    {
      if (!IsRequestAppend)
        // throws NotSupportedException
        return base.EndGetRequestStream(asyncResult);

      if (asyncResult != appendAsyncResult)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      return appendMessageBodyStream;
    }

    private ImapWebResponse GetAppendResponse()
    {
      if (appendAsyncResult == null)
        throw new InvalidOperationException("GetRequestStream not called");

      ImapAppendedUidSet appendedUid;

      var result = Session.EndAppend(appendAsyncResult, out appendedUid);

      if (result.Failed)
        return new ImapWebResponse(result);

      appendMessageBodyStream.Dispose();

      var response = new ImapWebResponse(result);

      if (appendedUid != null)
        response.SetResponseUri(new Uri(string.Format("{0}/;UID={1}", RequestUri, appendedUid.ToArray()[0])));

      return response;
    }

    private IAsyncResult appendAsyncResult = null;
    private ImapAppendMessageBodyStream appendMessageBodyStream = null;
#endregion

    private ImapWebResponse GetCreateResponse()
    {
      ImapMailbox createdMailbox;

      var result = Session.Create(RequestMailbox, out createdMailbox);

      if (result.Failed)
        return new ImapWebResponse(result);

      var response = new ImapWebResponse(result);

      response.Mailboxes = new[] {createdMailbox};

      Session.Subscribe(createdMailbox);
      Session.Lsub(createdMailbox);

      return response;
    }

    private ImapWebResponse GetDeleteResponse()
    {
      var mailbox = RequestMailbox;

      Session.Unsubscribe(mailbox);

      var result = Session.Delete(mailbox);

      if (result.Failed)
        return new ImapWebResponse(result);

      var response = new ImapWebResponse(result);

      return response;
    }

    private ImapWebResponse GetExpungeResponse()
    {
      var response = SelectRequestMailbox();

      if (response != null)
        return response;

      return new ImapWebResponse(Session.Expunge());
    }

    private ImapWebResponse GetExamineResponse()
    {
      return new ImapWebResponse(Session.Examine(RequestMailbox));
    }

    private ImapWebResponse GetRenameResponse()
    {
      ImapMailbox renamedMailbox;

      var mailbox = RequestMailbox;
      var result = Session.Rename(mailbox, GetDestinationMailbox(), out renamedMailbox);

      if (result.Failed)
        return new ImapWebResponse(result);

      Session.Unsubscribe(mailbox);
      Session.Subscribe(renamedMailbox);
      Session.Lsub(renamedMailbox);

      var response = new ImapWebResponse(result);

      response.Mailboxes = new[] {renamedMailbox};

      return response;
    }

    private ImapWebResponse GetSubscribeResponse()
    {
      return new ImapWebResponse(Session.Subscribe(RequestMailbox));
    }

    private ImapWebResponse GetUnsubscribeResponse()
    {
      return new ImapWebResponse(Session.Unsubscribe(RequestMailbox));
    }
  }
}
