// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Text;
using System.Threading;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.WebClient {
  public abstract class ImapWebRequest : WebRequest, IImapSessionProfile {
    private ImapSessionManager sessionManager;

    private ImapSession session = null;

    protected ImapSession Session {
      get { return session; }
    }

    private ICredentialsByHost credentials = null;

    /// <value>value must implement ICredentialsByHost</value>
    public override ICredentials Credentials {
      get { return credentials as ICredentials; }
      set
      {
        if (value != null && !(value is ICredentialsByHost))
          throw new ArgumentException("value must be ICredentialsByHost");
        credentials = value as ICredentialsByHost;
      }
    }

    ICredentialsByHost IImapSessionProfile.Credentials {
      get { return credentials; }
    }

    private Uri requestUri;

    public override Uri RequestUri {
      get { return requestUri; }
    }

    protected string RequestMailbox {
      get { return ImapStyleUriParser.GetDecodedMailbox(requestUri); }
    }

    Uri IImapSessionProfile.Authority {
      get { return new Uri(requestUri.GetLeftPart(UriPartial.Authority)); }
    }

    private int timeout = System.Threading.Timeout.Infinite;

    public override int Timeout {
      get { return timeout; }
      set
      {
        if (value < -1)
          throw new ArgumentOutOfRangeException("must be greater than or equals to -1");
        timeout = value;
      }
    }

    public override string Method {
      get; set;
    }

    /// <value>be used when method is COPY or RENAME</value>
    public Uri DestinationUri {
      get; set;
    }

    /// <value>be used when method is STORE</value>
    public ImapStoreDataItem StoreDataItem {
      get; set;
    }

    /// <value>be used when method is SORT</value>
    public ImapSortCriteria SortCriteria {
      get; set;
    }

    /// <value>be used when method is THREAD</value>
    public ImapThreadingAlgorithm ThreadingAlgorithm {
      get; set;
    }

    /// <value>be used when method is APPEND</value>
    public override long ContentLength {
      get; set;
    }

    public bool AllowLoginRedirect {
      get; set;
    }

    public bool AllowCreateMailbox {
      get; set;
    }

    public bool UseTlsIfAvailable {
      get; set;
    }

    public bool UseDeflateIfAvailable {
      get; set;
    }

    public bool KeepAlive {
      get; set;
    }

    public bool ReadOnly {
      get; set;
    }

    public ImapResponseCode[] ExpectedErrorResponseCodes {
      get; set;
    }

    public Encoding SearchCriteriaEncoding {
      get; set;
    }

    protected ImapWebRequest(Uri requestUri, ImapSessionManager sessionManager)
      : base()
    {
      // scheme is checked by IWebRequestCreate.Create
      this.requestUri = requestUri;

      this.sessionManager = sessionManager;
      this.Method = null;
      this.DestinationUri = null;
      this.ContentLength = 0;
      this.StoreDataItem = null;
      this.SortCriteria = null;
      this.ThreadingAlgorithm = null;
      this.AllowLoginRedirect = true;
      this.AllowCreateMailbox = false;
      this.UseTlsIfAvailable = true;
      this.UseDeflateIfAvailable = false;
      this.KeepAlive = true;
      this.ReadOnly = false;
      this.ExpectedErrorResponseCodes = new ImapResponseCode[] {};
      this.SearchCriteriaEncoding = Encoding.UTF8;
    }

    public override WebResponse GetResponse()
    {
      return EndGetResponse(BeginGetResponse(null, null));
    }

    private delegate ImapWebResponse GetResponseDelegate();

    public override IAsyncResult BeginGetResponse(AsyncCallback callback, object state)
    {
      return (new GetResponseDelegate(GetResponseProc)).BeginInvoke(null, null);
    }

    public override WebResponse EndGetResponse(IAsyncResult asyncResult)
    {
      var ar = asyncResult as System.Runtime.Remoting.Messaging.AsyncResult;

      if (ar == null)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      return (ar.AsyncDelegate as GetResponseDelegate).EndInvoke(ar);
    }

    protected abstract ImapWebResponse InternalGetResponse();

    private ImapWebResponse GetResponseProc()
    {
      try {
        GetSession();

        var response = InternalGetResponse();

        if (response.ResponseUri == null)
          response.SetResponseUri(requestUri);

        response.Session = session;
        response.SessionManager = sessionManager;
        response.KeepAlive = KeepAlive;

        if (response.Result.Succeeded)
          // succeeded
          return response;

        if (ExpectedErrorResponseCodes != null) {
          foreach (var code in ExpectedErrorResponseCodes) {
            var resp = response.Result.GetResponseCode(code);

            if (resp is ImapTaggedStatusResponse)
              // expected error
              return response;
          }
        }

        // unexpected error
        throw new WebException(response.ResponseDescription, null, WebExceptionStatus.ProtocolError, response);
      }
      finally {
        session = null;
      }
    }

    protected Uri GetDestinationUri(ImapUriForm requiredForm)
    {
      if (ImapStyleUriParser.GetAuthority(DestinationUri) != ImapStyleUriParser.GetAuthority(requestUri))
        throw new InvalidOperationException("destination must be same authority");

      var form = ImapStyleUriParser.GetUriForm(DestinationUri);

      if (form != requiredForm)
        throw new InvalidOperationException(string.Format("invalid form of destination URI, required form is {0} but was {1}", requiredForm, form));

      return DestinationUri;
    }

    protected string GetDestinationMailbox()
    {
      return ImapStyleUriParser.GetDecodedMailbox(GetDestinationUri(ImapUriForm.ListMessages));
    }

    protected ImapSession GetSession()
    {
      var s = sessionManager.GetExistSession(requestUri);

      if (s != null && !s.IsTransactionProceeding && s.NoOp().Failed && s.State == ImapSessionState.NotConnected) {
        // disconnected
        sessionManager.UnregisterSession(requestUri);
        s = null;
      }

      if (s == null) {
        var session = sessionManager.CreateSession(this);

        if (KeepAlive)
          sessionManager.RegisterSession(session);

        requestUri = new Uri(session.Authority, requestUri.PathAndQuery);

        s = session;
      }

      this.session = s;

      return s;
    }

    protected ImapWebResponse SelectRequestMailbox()
    {
      var mailbox = RequestMailbox;

      if (Session.State == ImapSessionState.Selected && !string.Equals(Session.SelectedMailbox.Name, mailbox)) {
        var resp = UnselectMailbox();

        if (resp != null)
          return resp;
      }

      if (Session.State != ImapSessionState.Selected) {
        var result = ReadOnly ? Session.Examine(mailbox) : Session.Select(mailbox);

        if (result.Failed)
          return new ImapWebResponse(result);
      }

      return null;
    }

    protected ImapWebResponse UnselectMailbox()
    {
      if (Session.State != ImapSessionState.Selected)
        return null;

      if (Session.ServerCapabilities.Has(ImapCapability.Unselect)) {
        var result = Session.Unselect();

        if (result.Failed)
          return new ImapWebResponse(result);
      }
      else {
        var result = Session.Close();

        if (result.Failed)
          return new ImapWebResponse(result);
      }

      return null;
    }
  }
}
