using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Tests;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionTest {
    [SetUp]
    public void Setup()
    {
      server = new ImapSimulatedServer();
      server.Start();

      host = server.ServerEndPoint.Address.ToString();
      port = server.ServerEndPoint.Port;

      credential = new NetworkCredential(username, password, host);
      uri = new Uri(string.Format("{0}://{1}@{2}:{3}",
                                  ImapUri.UriSchemeImap,
                                  username,
                                  host,
                                  port));
    }

    [TearDown]
    public void TearDown()
    {
      server.EnqueueResponse("* BYE Logging out\r\n");
      server.Stop();
    }

    private void AssertEqualCaseInsensitive(string expected, string actual)
    {
      AssertEqualCaseInsensitive(expected, actual, "sent command");
    }

    private void AssertEqualCaseInsensitive(string expected, string actual, string message)
    {
      Assert.IsTrue(string.Equals(expected, actual, StringComparison.InvariantCultureIgnoreCase),
                    message + "\r\nexpected: " + expected.Replace("\r", "\\r").Replace("\n", "\\n") + "\r\nactual: " + actual.Replace("\r", "\\r").Replace("\n", "\\n"));
    }

    [Test]
    public void TestConnect()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsFalse(session.IsIdling);
        Assert.IsNull(session.SelectedMailbox);
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        Assert.AreEqual(new Uri(string.Format("imap://{0}:{1}/", host, port)), session.Authority);
        Assert.AreEqual(1, session.ServerCapabilities.Count);
      }
    }

    [Test]
    public void TestImap4Rev1Incapable()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        server.EnqueueResponse("* CAPABILITY NONE\r\n0000 OK ImapCapability completed\r\n");

        ImapCapabilityList caps;

        Assert.IsTrue((bool)session.Capability(out caps));
        Assert.IsNotNull(caps);
        Assert.IsFalse(caps.Has(ImapCapability.Imap4Rev1));
        AssertEqualCaseInsensitive("0000 capability\r\n",
                                   server.DequeueRequest());

        try {
          session.HandlesIncapableAsException = true;
          session.NoOp();
          Assert.Fail("exception not occured");
        }
        catch (ImapIncapableException) {
        }
      }
    }

    [Test]
    public void TestGreetingWithCapability()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 AUTH=PLAIN CHILDREN THREAD=REFERENCES X-EXTENSION1 X-EXTENSION2] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        Assert.AreEqual(7, session.ServerCapabilities.Count);

        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.AuthCRAMMD5));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.AuthPlain));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Children));

        Assert.IsTrue(session.ServerCapabilities.Has("THREAD=REFERENCES"));
        Assert.IsTrue(session.ServerCapabilities.Has("X-EXTENSION1"));
        Assert.IsTrue(session.ServerCapabilities.Has("X-EXTENSION2"));
      }
    }

    [Test]
    public void TestGreetingPreAuthWithCapability()
    {
      server.EnqueueResponse("* PREAUTH [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.AreEqual(2, session.ServerCapabilities.Count);
        Assert.AreEqual(new Uri(string.Format("imap://{0}:{1}/", host, port)), session.Authority);

        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.AuthPlain));
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestGreetingOkWithNoImapCapability()
    {
      server.EnqueueResponse("* OK [CAPABILITY AUTH=PLAIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;

        Assert.IsTrue(!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1));

        session.HandlesIncapableAsException = true;

        session.Login();
      }
    }

    [Test]
    public void TestGreetingHomeServerReferralAsException()
    {
      server.EnqueueResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n");

      try {
        new ImapSession(host, port, true);

        Assert.Fail("connected without exception");
      }
      catch (ImapLoginReferralException ex) {
        Assert.IsTrue(ex.Message.Contains("Server not accepting connections.  Try SERVER2"));
        Assert.AreEqual(new Uri("IMAP://user;AUTH=*@SERVER2/"), ex.ReferToUri);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapException))]
    public void TestGreetingHomeServerReferralAsError()
    {
      server.EnqueueResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n");

      using (var session = new ImapSession(host, port)) {
      }
    }

    [Test]
    public void TestLogin()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;
        session.HandlesIncapableAsException = true;

        server.EnqueueResponse("0000 NO incorrect\r\n");

        Assert.IsFalse((bool)session.Login());

        AssertEqualCaseInsensitive(string.Format("0000 login {0} {1}\r\n", username, password),
                                   server.DequeueRequest());

        server.EnqueueResponse("0001 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Login());

        AssertEqualCaseInsensitive(string.Format("0001 login {0} {1}\r\n", username, password),
                                   server.DequeueRequest());

        Assert.AreEqual(new Uri(string.Format("imap://{0}@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestLoginDisabled()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 LOGINDISABLED] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;
        session.Login();
      }
    }

    [Test]
    public void TestLoginHomerServerReferralAsException()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, true)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 NO [REFERRAL IMAP://MIKE@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

        try {
          session.Login();

          Assert.Fail("logged in without exception");
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Specified user is invalid on this server. Try SERVER2."));
          Assert.AreEqual(new Uri("IMAP://MIKE@SERVER2/"), ex.ReferToUri);
        }
      }
    }

    [Test]
    public void TestLoginHomerServerReferralAsError()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, false)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 NO [REFERRAL IMAP://MIKE@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

        Assert.IsFalse((bool)session.Login());
      }
    }

    [Test]
    public void TestLoginMailboxReferralAsException()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, true)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 OK [REFERRAL IMAP://MATTHEW@SERVER2/] Specified user's personal mailboxes located on Server2, but public mailboxes are available.\r\n");

        try {
          session.Login();

          Assert.Fail("logged in without exception");
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Specified user's personal mailboxes located on Server2, but public mailboxes are available."));
          Assert.AreEqual(new Uri("IMAP://MATTHEW@SERVER2/"), ex.ReferToUri);
        }
      }
    }

    [Test]
    public void TestLoginMailboxReferralAsError()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, false)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 OK [REFERRAL IMAP://MATTHEW@SERVER2/] Specified user's personal mailboxes located on Server2, but public mailboxes are available.\r\n");

        Assert.IsTrue((bool)session.Login());
      }
    }

    [Test]
    public void TestAuthenticateMailboxReferralAsException()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN LOGIN-REFERRALS] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port, true)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 NO [REFERRAL IMAP://user;AUTH=GSSAPI@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

        try {
          session.Authenticate(ImapAuthenticationMechanism.Login);

          Assert.Fail("authentcated without exception");
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Specified user is invalid on this server. Try SERVER2."));
          Assert.AreEqual(new Uri("IMAP://user;AUTH=GSSAPI@SERVER2/"), ex.ReferToUri);
        }
      }
    }

    [Test]
    public void TestAuthenticateLogin()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;

        server.EnqueueResponse("+ " + Convert.ToBase64String(latin1Encoding.GetBytes("Username:")) + "\r\n");
        server.EnqueueResponse("+ " + Convert.ToBase64String(latin1Encoding.GetBytes("Password:")) + "\r\n");
        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(ImapAuthenticationMechanism.Login));

        AssertEqualCaseInsensitive("0000 authenticate login\r\n",
                                   server.DequeueRequest());

        var requested = server.DequeueRequest();

        AssertEqualCaseInsensitive(username,
                                   latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

        requested = server.DequeueRequest();

        AssertEqualCaseInsensitive(password,
                                   latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

        Assert.AreEqual(new Uri(string.Format("imap://{0};AUTH=LOGIN@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test]
    public void TestAuthenticatePlain()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;

        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(ImapAuthenticationMechanism.Plain));

        AssertEqualCaseInsensitive("0000 authenticate plain\r\n",
                                   server.DequeueRequest());

        var requested = server.DequeueRequest();
        var userpass = latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))).Split(new[] {"\0"}, StringSplitOptions.RemoveEmptyEntries);

        Assert.AreEqual(credential.Domain, userpass[0]);
        Assert.AreEqual(credential.UserName, userpass[1]);
        Assert.AreEqual(credential.Password, userpass[2]);

        Assert.AreEqual(new Uri(string.Format("imap://{0};AUTH=PLAIN@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test]
    public void TestAuthenticateCramMd5()
    {
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5] ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;

        var timestamp = ((long)DateTime.Now.Subtract(new DateTime(1970, 1, 1)).TotalMilliseconds).ToString();
        var challenge = latin1Encoding.GetBytes(string.Format("<{0}@{1}>", timestamp, host));

        server.EnqueueResponse("+ " + Convert.ToBase64String(challenge) + "\r\n");
        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Authenticate(ImapAuthenticationMechanism.CRAMMD5));

        AssertEqualCaseInsensitive("0000 authenticate cram-md5\r\n",
                                   server.DequeueRequest());

        var requested = server.DequeueRequest();
        var userkeyed = latin1Encoding.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))).Split(' ');

        Assert.AreEqual(username, userkeyed[0]);

        var keyed = new List<byte>();

        for (var index = 0; index < userkeyed[1].Length; index += 2) {
          keyed.Add(Convert.ToByte(userkeyed[1].Substring(index, 2), 16));
        }

        Assert.AreEqual(Convert.ToBase64String((new HMACMD5(latin1Encoding.GetBytes(password))).ComputeHash(challenge)),
                        Convert.ToBase64String(keyed.ToArray()));

        Assert.AreEqual(new Uri(string.Format("imap://{0};AUTH=CRAM-MD5@{1}:{2}/", username, host, port)), session.Authority);
      }
    }

    [Test, Ignore("old feature")]
    public void TestSelectBestAuthentication()
    {
#if false
      server.EnqueueResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 AUTH=LOGIN AUTH=PLAIN] ImapSimulatedServer ready\r\n");

      var loginUri = new ImapUri(string.Format("imap://{0};AUTH=KERBEROS_V4@{1}:{2}", uri.UserName, uri.Host, uri.Port));

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 NO authentication failed\r\n");
        server.EnqueueResponse("0001 NO authentication failed\r\n");
        server.EnqueueResponse("0002 NO authentication failed\r\n");
        server.EnqueueResponse("0003 NO authentication failed\r\n");

        Assert.IsFalse(session.Authenticate());

        //AssertEqualCaseInsensitive("0000 authenticate kerberos_v4\r\n",
        //                           server.DequeueRequest());
        AssertEqualCaseInsensitive("0000 authenticate cram-md5\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0001 authenticate plain\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0002 authenticate login\r\n",
                                   server.DequeueRequest());
      }
#endif
    }

    [Test]
    public void TestLogoutNotAuthenticated()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");
      server.EnqueueResponse("* BYE Logging out\r\n");

      using (var session = new ImapSession(host, port)) {
        Assert.IsTrue(session.Logout().Code == Transaction.ImapTransactionResultCode.Bye);

        Assert.IsTrue(ImapSessionState.NotConnected == session.State ||
                      ImapSessionState.NotAuthenticated == session.State);

        Assert.AreEqual(null, session.Authority);
      }
    }

    [Test]
    public void TestLogoutAuthenticated()
    {
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 OK authenticated\r\n");

        Assert.IsTrue((bool)session.Login());

        Assert.AreEqual(new Uri(string.Format("imap://{0}@{1}:{2}/", username, host, port)), session.Authority);

        server.DequeueRequest();

        server.EnqueueResponse("* BYE Logging out\r\n");

        Assert.IsTrue(session.Logout().Code == Transaction.ImapTransactionResultCode.Bye);

        AssertEqualCaseInsensitive("0001 logout\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue(ImapSessionState.NotConnected == session.State ||
                      ImapSessionState.NotAuthenticated == session.State);

        Assert.AreEqual(null, session.Authority);
      }
    }

    [Test]
    public void TestListStatusSelect()
    {
      // greeting and login transaction
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      using (var session = new ImapSession(host, port)) {
        session.Credentials = credential;

        server.EnqueueResponse("0000 OK authenticated\r\n");
  
        session.Login();

        Assert.AreEqual(new Uri(string.Format("imap://{0}@{1}:{2}/", username, host, port)), session.Authority);

        server.DequeueRequest();
  
        // CAPABILITY transaction
        server.EnqueueResponse("* CAPABILITY IMAP4rev1\r\n" +
                               "0001 OK CAPABILITY completed\r\n");
  
        session.Capability();
  
        server.DequeueRequest();
  
        // LIST transaction
        server.EnqueueResponse("* LIST (\\HasChildren) \".\" \"INBOX\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        Assert.IsTrue((bool)session.List("*", out mailboxes));
  
        AssertEqualCaseInsensitive("0002 list \"\" \"*\"\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.HasChildren));
        Assert.AreEqual(uri.ToString() + "INBOX/", mailboxes[0].Url.ToString());
  
        // LSUB transaction
        server.EnqueueResponse("* LSUB (\\HasChildren) \"/\" \"Trash\"\r\n" +
                               "* LSUB (\\HasNoChildren) \"/\" \"Trash/mail\"\r\n" +
                               "* LSUB (\\HasNoChildren) \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "* LSUB (\\HasNoChildren) \".\" \"INBOX.&U,BTFw-\"\r\n" +
                               "0003 OK LSUB completed\r\n");
  
        Assert.IsTrue((bool)session.Lsub("*", out mailboxes));
  
        AssertEqualCaseInsensitive("0003 lsub \"\" \"*\"\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(4, mailboxes.Length);
        Assert.AreEqual("Trash", mailboxes[0].Name);
        Assert.IsTrue(mailboxes[0].Flags.Has(ImapMailboxFlag.HasChildren));
  
        Assert.AreEqual("Trash/mail", mailboxes[1].Name);
        Assert.IsTrue(mailboxes[1].Flags.Has(ImapMailboxFlag.HasNoChildren));
        Assert.AreEqual(uri.ToString() + "Trash/mail/", mailboxes[1].Url.ToString());
  
        Assert.AreEqual("INBOX.日本語", mailboxes[2].Name);
        Assert.AreEqual("INBOX.台北", mailboxes[3].Name);
  
        // STATUS transaction
        server.EnqueueResponse("* STATUS \"Trash\" (MESSAGES 3 RECENT 1 UIDVALIDITY 123456 UNSEEN 2 UIDNEXT 4)\r\n" +
                               "0004 OK STATUS completed\r\n");
  
        Assert.IsTrue((bool)session.Status(mailboxes[0], ImapStatusDataItem.StandardAll));
  
        AssertEqualCaseInsensitive("0004 status \"Trash\" (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(3, mailboxes[0].ExistsMessage);
        Assert.AreEqual(1, mailboxes[0].RecentMessage);
        Assert.AreEqual(123456, mailboxes[0].UidValidity);
        Assert.AreEqual(2, mailboxes[0].UnseenMessage);
        Assert.AreEqual(4, mailboxes[0].UidNext);
  
        // SELECT transaction
        Assert.IsNull(session.SelectedMailbox);
  
        server.EnqueueResponse("* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                               "* 15 EXISTS\r\n" +
                               "* 2 RECENT\r\n" +
                               "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                               "* OK [UIDNEXT 16]\r\n" +
                               "* OK [UNSEEN 13]\r\n" +
                               "0005 OK [READ-WRITE] SELECT completed\r\n");
  
        var mailbox = mailboxes[2];
  
        Assert.IsTrue((bool)session.Select(mailbox));
  
        AssertEqualCaseInsensitive("0005 select \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.AreSame(mailbox, session.SelectedMailbox);
  
        Assert.AreEqual("INBOX.日本語", mailbox.Name);
        Assert.AreEqual(15, mailbox.ExistsMessage);
        Assert.AreEqual(2, mailbox.RecentMessage);
        Assert.AreEqual(0, mailbox.UnseenMessage);
        Assert.AreEqual(13, mailbox.FirstUnseen);
        Assert.AreEqual(1202674433, mailbox.UidValidity);
        Assert.AreEqual(16, mailbox.UidNext);
        Assert.IsFalse(mailbox.ReadOnly);
        Assert.IsTrue(mailbox.UidPersistent);

        /*
        Assert.AreEqual(selectedMailbox.Name, mailbox.Name);
        Assert.AreEqual(selectedMailbox.ExistsMessage, mailbox.ExistsMessage);
        Assert.AreEqual(selectedMailbox.RecentMessage, mailbox.RecentMessage);
        Assert.AreEqual(selectedMailbox.UnseenMessage, mailbox.UnseenMessage);
        Assert.AreEqual(selectedMailbox.FirstUnseen, mailbox.FirstUnseen);
        Assert.AreEqual(selectedMailbox.UidValidity, mailbox.UidValidity);
        Assert.AreEqual(selectedMailbox.UidNext, mailbox.UidNext);
        Assert.AreEqual(selectedMailbox.ReadOnly, mailbox.ReadOnly);
        Assert.AreEqual(selectedMailbox.ApplicableFlags.Count, mailbox.ApplicableFlags.Count);
        Assert.AreEqual(selectedMailbox.PermanentFlags.Count, mailbox.PermanentFlags.Count);
        */

        Assert.AreEqual(7, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.ApplicableFlags.Has(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.ApplicableFlags.Has("custom1"));
        Assert.IsTrue(mailbox.ApplicableFlags.Has("custom2"));
  
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.PermanentFlags.Has(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.PermanentFlags.Has("custom1"));
  
        // CLOSE transaction
        server.EnqueueResponse("0006 OK CLOSE completed\r\n");
  
        Assert.IsTrue((bool)session.Close());
  
        AssertEqualCaseInsensitive("0006 close\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      }
    }

    private ImapSession SelectMailbox(params string[] additionalCapabilities)
    {
      // greeting and login transaction
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      var session = new ImapSession(host, port);

      session.Credentials = credential;
      session.HandlesIncapableAsException = true;

      server.EnqueueResponse("0000 OK authenticated\r\n");

      session.Login();

      Assert.AreEqual(new Uri(string.Format("imap://{0};AUTH=LOGIN@{1}:{2}/", username, host, port)), session.Authority);

      server.DequeueRequest();

      // CAPABILITY transaction
      server.EnqueueResponse(string.Format("* CAPABILITY IMAP4rev1 {0}\r\n", string.Join(" ", additionalCapabilities)) +
                             "0001 OK CAPABILITY completed\r\n");

      session.Capability();

      server.DequeueRequest();

      // LIST transaction
      server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                             "0002 OK LIST completed\r\n");

      ImapMailbox[] mailboxes;

      session.List("", out mailboxes);

      server.DequeueRequest();

      // SELECT transaction
      server.EnqueueResponse("* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                             "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                             "* 15 EXISTS\r\n" +
                             "* 2 RECENT\r\n" +
                             "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                             "* OK [UIDNEXT 16]\r\n" +
                             "* OK [UNSEEN 13]\r\n" +
                             "0003 OK [READ-WRITE] SELECT completed\r\n");

      session.Select(mailboxes[0]);

      server.DequeueRequest();

      return session;
    }

    private void CloseMailbox(ImapSession session)
    {
      CloseMailbox(session, "0005");
    }

    private void CloseMailbox(ImapSession session, string tag)
    {
      // CLOSE transaction
      server.EnqueueResponse(tag + " OK CLOSE completed\r\n");

      session.Close();

      AssertEqualCaseInsensitive(tag + " close\r\n",
                                 server.DequeueRequest());

      session.Disconnect(false);
    }

    [Test]
    public void TestSelectUidNotSticky()
    {
      using (var session = Authenticate()) {
        // SELECT transaction
        server.EnqueueResponse("* 1 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 1] Message 1 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] Validity session-only\r\n" +
                               "* OK [UIDNEXT 2] Predicted next UID\r\n" +
                               "* NO [UIDNOTSTICKY] Non-persistent UIDs\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen)] Limited\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed\r\n");

        Assert.IsTrue((bool)session.Select("funny"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"funny\"\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.SelectedMailbox.UidPersistent);
        Assert.AreEqual(2, session.SelectedMailbox.UidNext);
        Assert.AreEqual(3857529045, session.SelectedMailbox.UidValidity);
        Assert.IsFalse(session.SelectedMailbox.ModificationSequences);
      }
    }

    [Test]
    public void TestSelectHighestModSeq()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // SELECT transaction
        server.EnqueueResponse("* 172 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                               "* OK [HIGHESTMODSEQ 715194045007]\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed\r\n");

        Assert.IsTrue((bool)session.Select("funny"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"funny\"\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue(session.SelectedMailbox.ModificationSequences);
        Assert.AreEqual(715194045007UL, session.SelectedMailbox.HighestModSeq);
      }
    }

    [Test]
    public void TestSelectNoModSeq()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // SELECT transaction
        server.EnqueueResponse("* 172 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                               "* OK [NOMODSEQ] Sorry, this mailbox format doesn't support modsequences\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed\r\n");

        Assert.IsTrue((bool)session.Select("funny"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"funny\"\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(session.SelectedMailbox.ModificationSequences);
        Assert.AreEqual(0UL, session.SelectedMailbox.HighestModSeq);
      }
    }

    [Test]
    public void TestSelectCondStore()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // SELECT transaction
        server.EnqueueResponse("* 172 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                               "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                               "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                               "* OK [HIGHESTMODSEQ 715194045007]\r\n" +
                               "0002 OK [READ-WRITE] SELECT completed, CONDSTORE is now enabled\r\n");

        Assert.IsTrue((bool)session.SelectCondStore("INBOX"));
  
        AssertEqualCaseInsensitive("0002 SELECT \"INBOX\" (CONDSTORE)\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue((bool)session.SelectedMailbox.ModificationSequences);
        Assert.AreEqual(715194045007UL, session.SelectedMailbox.HighestModSeq);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestSelectCondStoreIncapable()
    {
      using (var session = Authenticate()) {
        session.HandlesIncapableAsException = true;

        // SELECT transaction
        session.SelectCondStore("INBOX");
      }
    }

    [Test]
    public void TestNoOp()
    {
      using (var session = SelectMailbox()) {
        // NOOP transaction
        server.EnqueueResponse("0004 OK NOOP completed\r\n");
  
        Assert.IsTrue((bool)session.NoOp());
  
        AssertEqualCaseInsensitive("0004 noop\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestNoOpStatusUpdate()
    {
      using (var session = SelectMailbox()) {
        // NOOP transaction
        server.EnqueueResponse("* 22 EXPUNGE\r\n" +
                               "* 18 EXPUNGE\r\n" +
                               "* 23 EXISTS\r\n" +
                               "* 3 RECENT\r\n" +
                               "* 14 FETCH (FLAGS (\\Seen \\Deleted))\r\n" +
                               "0004 OK NOOP completed\r\n");
  
        Assert.IsTrue((bool)session.NoOp());
  
        AssertEqualCaseInsensitive("0004 noop\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapUpdatedStatusKind.Expunged | ImapUpdatedStatusKind.Exists | ImapUpdatedStatusKind.Recent | ImapUpdatedStatusKind.Messages,
                        session.LastUpdatedStatus.UpdatedKind);
        Assert.AreEqual(1, session.LastUpdatedStatus.MessageAttributes.Length);
        Assert.IsTrue(session.LastUpdatedStatus.MessageAttributes[0].Flags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(session.LastUpdatedStatus.MessageAttributes[0].Flags.Has(ImapMessageFlag.Deleted));
  
        Assert.AreEqual(2, session.LastUpdatedStatus.Expunged.Length);
        Assert.AreEqual(22, session.LastUpdatedStatus.Expunged[0]);
        Assert.AreEqual(18, session.LastUpdatedStatus.Expunged[1]);
  
        Assert.IsNotNull(session.LastUpdatedStatus.Exists);
        Assert.AreEqual(23, (long)session.LastUpdatedStatus.Exists);
  
        Assert.IsNotNull(session.LastUpdatedStatus.Recent);
        Assert.AreEqual(3, (long)session.LastUpdatedStatus.Recent);
  
        Assert.AreEqual(23, session.SelectedMailbox.ExistsMessage, "selected mailbox updated");
        Assert.AreEqual(3, session.SelectedMailbox.RecentMessage, "selected mailbox updated");
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestCheckStatusUpdate()
    {
      using (var session = SelectMailbox()) {
        // CHECK transaction
        server.EnqueueResponse("* 23 EXISTS\r\n" +
                               "0004 OK CHECK completed\r\n");

        Assert.IsTrue((bool)session.Check());
  
        AssertEqualCaseInsensitive("0004 check\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(0, session.LastUpdatedStatus.MessageAttributes.Length);
        Assert.IsNotNull(session.LastUpdatedStatus.Expunged);
        Assert.AreEqual(0, session.LastUpdatedStatus.Expunged.Length);
        Assert.IsNotNull(session.LastUpdatedStatus.Exists);
        Assert.AreEqual(23, (long)session.LastUpdatedStatus.Exists);
        Assert.IsNull(session.LastUpdatedStatus.Recent);
  
        Assert.AreEqual(23, session.SelectedMailbox.ExistsMessage, "selected mailbox updated");
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestStatus()
    {
      using (var session = SelectMailbox()) {
        // STATUS transaction
        server.EnqueueResponse("* STATUS \"INBOX\" (UIDNEXT 4 UNSEEN 2)\r\n" +
                               "0004 OK STATUS completed\r\n");
  
        ImapStatusAttributeList statusAttributes;
  
        Assert.IsTrue((bool)session.Status("INBOX", ImapStatusDataItem.UidNext + ImapStatusDataItem.Unseen, out statusAttributes));

        Assert.IsNull(statusAttributes.Messages);
        Assert.IsNull(statusAttributes.Recent);
        Assert.IsNull(statusAttributes.UidValidity);

        Assert.IsNotNull(statusAttributes.UidNext);
        Assert.AreEqual(4, statusAttributes.UidNext);
        Assert.IsNotNull(statusAttributes.Unseen);
        Assert.AreEqual(2, statusAttributes.Unseen);
  
        Assert.AreEqual(4, session.SelectedMailbox.UidNext);
        Assert.AreEqual(2, session.SelectedMailbox.UnseenMessage);

        AssertEqualCaseInsensitive("0004 status \"INBOX\" (UIDNEXT UNSEEN)\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestStatusHighestModSeq()
    {
      using (var session = Authenticate("CONDSTORE")) {
        // STATUS transaction
        server.EnqueueResponse("* STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292 HIGHESTMODSEQ 7011231777)\r\n" +
                               "0002 OK STATUS completed\r\n");
  
        ImapStatusAttributeList statusAttributes;
  
        Assert.IsTrue((bool)session.Status("blurdybloop", ImapStatusDataItem.Messages + ImapStatusDataItem.UidNext + ImapStatusDataItem.HighestModSeq, out statusAttributes));

        Assert.IsNull(statusAttributes.Recent);
        Assert.IsNull(statusAttributes.Unseen);
        Assert.IsNull(statusAttributes.UidValidity);

        Assert.IsNotNull(statusAttributes.Messages);
        Assert.AreEqual(231, statusAttributes.Messages);
        Assert.IsNotNull(statusAttributes.UidNext);
        Assert.AreEqual(44292, statusAttributes.UidNext);
        Assert.IsNotNull(statusAttributes.HighestModSeq);
        Assert.AreEqual(7011231777UL, statusAttributes.HighestModSeq);

        AssertEqualCaseInsensitive("0002 STATUS \"blurdybloop\" (MESSAGES UIDNEXT HIGHESTMODSEQ)\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestFetchEmptySequenceSet()
    {
      using (var session = SelectMailbox()) {
        ImapMessage[] messages;

        session.Fetch(ImapSequenceSet.CreateSet(new long[] {}), ImapFetchDataItem.Fast, out messages);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestFetchEmptyUidSet()
    {
      using (var session = SelectMailbox()) {
        ImapMessage[] messages;

        session.Fetch(ImapSequenceSet.CreateUidSet(new long[] {}), ImapFetchDataItem.Fast, out messages);
      }
    }

    [Test]
    public void TestFetchMessageInfo()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (" +
                               "UID 4 " + 
                               "FLAGS (\\Answered custom1) " +
                               "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                               "RFC822.SIZE 1716" + 
                               ")\r\n" + 
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
        ImapFetchDataItem dataItem =
          ImapFetchDataItem.Uid +
          ImapFetchDataItem.Flags +
          ImapFetchDataItem.InternalDate +
          ImapFetchDataItem.Rfc822Size;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), dataItem, out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 (uid flags internaldate rfc822.size)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
  
        ImapMessage message = messages[0];
  
        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
        Assert.AreEqual(new Uri(session.SelectedMailbox.Url, ";UID=4"), message.Url);
        Assert.AreEqual(4, message.Uid);
        Assert.AreEqual(1, message.Sequence);
        Assert.IsTrue(message.Flags.Has(ImapMessageFlag.Answered));
        Assert.IsTrue(message.Flags.Has("custom1"));
        Assert.AreEqual(new DateTimeOffset(2008, 2, 19, 8, 15, 48, new TimeSpan(+9, 0, 0)), message.InternalDate);
        Assert.AreEqual(1716, message.Rfc822Size);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchEnvelope()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (ENVELOPE (" + 
                               "\"Tue, 19 Feb 2008 08:15:48 +0900\" " +
                               "\"fetch test\" " + 
                               "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                               "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                               "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                               "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                               "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                               "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                               "\"<in.reply.to@mail.example.com>\" " + 
                               "\"<message.id@mail.example.com>\" " + 
                               "))\r\n" +
                               "0004 OK FETCH completed\r\n");
    
        ImapMessage[] messages;
    
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.Envelope, out messages));
    
        AssertEqualCaseInsensitive("0004 fetch 1 (envelope)\r\n",
                                   server.DequeueRequest());
    
        Assert.AreEqual(1, messages.Length);
    
        ImapMessage message = messages[0];
    
        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
    
        // envelope
        ImapEnvelope envelope = message.Envelope;
        Assert.AreEqual("Tue, 19 Feb 2008 08:15:48 +0900", envelope.Date);
        Assert.AreEqual("fetch test", envelope.Subject);
        Assert.AreEqual(1, envelope.From.Length);
        Assert.AreEqual("from address", envelope.From[0].Name);
        Assert.AreEqual(1, envelope.Sender.Length);
        Assert.AreEqual("sender address", envelope.Sender[0].Name);
        Assert.AreEqual(1, envelope.ReplyTo.Length);
        Assert.AreEqual("reply-to address", envelope.ReplyTo[0].Name);
        Assert.AreEqual(1, envelope.To.Length);
        Assert.AreEqual("to address", envelope.To[0].Name);
        Assert.AreEqual(1, envelope.Cc.Length);
        Assert.AreEqual("cc address", envelope.Cc[0].Name);
        Assert.AreEqual(1, envelope.Bcc.Length);
        Assert.AreEqual("bcc address", envelope.Bcc[0].Name);
        Assert.AreEqual("<in.reply.to@mail.example.com>", envelope.InReplyTo);
        Assert.AreEqual("<message.id@mail.example.com>", envelope.MessageId);
    
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchBodyStructure()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY (\"text\" \"plain\" (\"charset\" \"ISO-2022-JP\") NIL NIL \"7bit\" 22 1))\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.Body, out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 (body)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
  
        ImapMessage message = messages[0];
  
        //Assert.AreEqual(session.SelectedMailbox, message.Mailbox);
  
        // body structure
        Assert.IsTrue(message.BodyStructure is ImapSinglePartBodyStructure);
  
        var structure = message.BodyStructure as ImapSinglePartBodyStructure;
  
        Assert.AreEqual("text", structure.Type);
        Assert.AreEqual("plain", structure.Subtype);
        Assert.AreEqual(1, structure.Parameters.Count);
        Assert.AreEqual("ISO-2022-JP", structure.Parameters["charset"]);
        Assert.IsNull(structure.Id);
        Assert.IsNull(structure.Description);
        Assert.AreEqual("7bit", structure.Encoding);
        Assert.AreEqual(22, structure.Size);
        Assert.AreEqual(1, structure.Lines);
        Assert.IsNull(structure.BodyStructure);
        Assert.IsNull(structure.Envelope);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchByMacro()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (" +
                               "FLAGS (\\Answered custom1) " +
                               "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                               "RFC822.SIZE 1716" + 
                               ")\r\n" + 
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.Fast, out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 FAST\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
        Assert.AreEqual(new DateTimeOffset(2008, 2, 19, 8, 15, 48, new TimeSpan(+9, 0, 0)), messages[0].InternalDate);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchBody()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY[] {12}\r\ntest message)\r\n" +
                               "* 2 FETCH (BODY[] \"test message\")\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateRangeSet(1, 2), ImapFetchDataItem.BodyPeek(), out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1:2 (body.peek[])\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(2, messages.Length);
  
        //Assert.AreEqual(session.SelectedMailbox, messages[0].Mailbox);
        //Assert.AreEqual(session.SelectedMailbox, messages[1].Mailbox);
  
        Assert.AreEqual("test message", Encoding.ASCII.GetString(messages[0].Body));
        Assert.AreEqual("test message", Encoding.ASCII.GetString(messages[1].Body));
  
        CloseMailbox(session);
      }
    }

    private string CreateLiteral(string body, Encoding encoding, int maxoctet)
    {
      var literalizedBody = encoding.GetBytes(body);

      var octetBody = (maxoctet == 0) ? literalizedBody.Length : maxoctet;

      return string.Format("{{{0}}}\r\n{1}",
                           octetBody,
                           latin1Encoding.GetString(literalizedBody, 0, octetBody));
    }

    [Test]
    public void TestFetchSection()
    {
      using (var session = SelectMailbox()) {
        var body = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 25 Feb 2008 15:01:12 +0900\r\n";
  
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)] " +
                               CreateLiteral(body, Encoding.ASCII, 0) +
                               ")\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), ImapFetchDataItem.BodyPeek("header.fields (from to subject date)"), out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 (body.peek[header.fields (from to subject date)])\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
  
        Assert.AreEqual(body, Encoding.ASCII.GetString(messages[0].Body));
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestFetchPartial()
    {
      using (var session = SelectMailbox()) {
        var body = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 25 Feb 2008 15:01:12 +0900\r\n";
        var octet = 10;
  
        // FETCH
        server.EnqueueResponse("* 1 FETCH (BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)] " +
                               CreateLiteral(body, Encoding.ASCII, octet) +
                               ")\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        var messageDataItems = ImapFetchDataItem.BodyText("header.fields (from to subject date)", 0, octet);
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateSet(1), messageDataItems, out messages));
  
        AssertEqualCaseInsensitive("0004 fetch 1 (body[header.fields (from to subject date)]<0." + octet.ToString() + ">)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, messages.Length);
  
        Assert.AreEqual(body.Substring(0, octet), Encoding.ASCII.GetString(messages[0].Body));
  
        CloseMailbox(session);
      }
    }


    [Test]
    public void TestFetchModSeq()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (MODSEQ (624140003))\r\n" +
                               "* 2 FETCH (MODSEQ (624140007))\r\n" +
                               "* 3 FETCH (MODSEQ (624140005))\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.Fetch(ImapSequenceSet.CreateRangeSet(1, 3), ImapFetchDataItem.ModSeq, out messages));
  
        AssertEqualCaseInsensitive("0004 FETCH 1:3 (MODSEQ)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(3, messages.Length);
        Assert.AreEqual(1, messages[0].Sequence);
        Assert.AreEqual(624140003UL, messages[0].ModSeq);
        Assert.AreEqual(2, messages[1].Sequence);
        Assert.AreEqual(624140007UL, messages[1].ModSeq);
        Assert.AreEqual(3, messages[2].Sequence);
        Assert.AreEqual(624140005UL, messages[2].ModSeq);
  
        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestFetchModSeqIncapable()
    {
      using (var session = SelectMailbox()) {
        session.HandlesIncapableAsException = true;

        // FETCH
        ImapMessage[] messages;

        session.Fetch(ImapSequenceSet.CreateRangeSet(1, 3), ImapFetchDataItem.Fast + ImapFetchDataItem.ModSeq, out messages);
      }
    }

    [Test]
    public void TestFetchChangedSince()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // FETCH
        server.EnqueueResponse("* 1 FETCH (UID 4 MODSEQ (65402) FLAGS (\\Seen))\r\n" +
                               "* 2 FETCH (UID 6 MODSEQ (75403) FLAGS (\\Deleted))\r\n" +
                               "* 4 FETCH (UID 8 MODSEQ (29738) FLAGS ($NoJunk $AutoJunk $MDNSent))\r\n" +
                               "0004 OK FETCH completed\r\n");
  
        ImapMessage[] messages;
  
        Assert.IsTrue((bool)session.FetchChangedSince(ImapSequenceSet.CreateUidFromSet(1), ImapFetchDataItem.Flags, 12345UL, out messages));
  
        AssertEqualCaseInsensitive("0004 UID FETCH 1:* (FLAGS) (CHANGEDSINCE 12345)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1, messages[0].Sequence);
        Assert.AreEqual(4, messages[0].Uid);
        Assert.AreEqual(65402UL, messages[0].ModSeq);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));

        Assert.AreEqual(2, messages[1].Sequence);
        Assert.AreEqual(6, messages[1].Uid);
        Assert.AreEqual(75403UL, messages[1].ModSeq);
        Assert.IsTrue(messages[1].Flags.Has(ImapMessageFlag.Deleted));

        Assert.AreEqual(4, messages[2].Sequence);
        Assert.AreEqual(8, messages[2].Uid);
        Assert.AreEqual(29738UL, messages[2].ModSeq);
        Assert.IsTrue(messages[2].Flags.Has("$NoJunk"));
        Assert.IsTrue(messages[2].Flags.Has("$AutoJunk"));
        Assert.IsTrue(messages[2].Flags.Has("$MDNSent"));

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestFetchChangedSinceIncapable()
    {
      using (var session = SelectMailbox()) {
        // FETCH
        ImapMessage[] messages;
        session.FetchChangedSince(ImapSequenceSet.CreateUidFromSet(1), ImapFetchDataItem.Flags, 12345UL, out messages);
      }
    }

    [Test]
    public void TestStore()
    {
      using (var session = SelectMailbox()) {
        // STORE
        server.EnqueueResponse("* 2 FETCH (FLAGS (\\Deleted \\Seen))\r\n" +
                               "* 3 FETCH (FLAGS (\\Deleted))\r\n" +
                               "* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Seen))\r\n" +
                               "0004 OK STORE completed\r\n");

        ImapMessage[] messages;

        Assert.IsTrue((bool)session.Store(ImapSequenceSet.CreateRangeSet(2, 4),
                                    ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted),
                                    out messages));

        AssertEqualCaseInsensitive("0004 STORE 2:4 +FLAGS (\\Deleted)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(2, messages[0].Sequence);
        Assert.AreEqual(0UL, messages[0].ModSeq);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messages[1].Sequence);
        Assert.AreEqual(0UL, messages[1].ModSeq);
        Assert.AreEqual(1, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Has(ImapMessageFlag.Deleted));

        Assert.AreEqual(4, messages[2].Sequence);
        Assert.AreEqual(0UL, messages[2].ModSeq);
        Assert.AreEqual(3, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Flagged));
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Seen));

        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestStoreEmptySequenceSet()
    {
      using (var session = SelectMailbox()) {
        session.Store(ImapSequenceSet.CreateSet(new long[] {}), ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted));
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestStoreEmptyUidSet()
    {
      using (var session = SelectMailbox()) {
        session.Store(ImapSequenceSet.CreateUidSet(new long[] {}), ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted));
      }
    }

    [Test]
    public void TestStoreUnchangedSince()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // STORE
        server.EnqueueResponse("* 7 FETCH (MODSEQ (320162342) FLAGS (\\Seen \\Deleted))\r\n" +
                               "* 5 FETCH (MODSEQ (320162350))\r\n" +
                               "* 9 FETCH (MODSEQ (320162349) FLAGS (\\Answered))\r\n" +
                               "0004 OK [MODIFIED 7,9] Conditional STORE failed\r\n");

        ImapMessage[] messages;
        ImapSequenceSet failedMessageSet;
  
        Assert.IsTrue((bool)session.StoreUnchangedSince(ImapSequenceSet.CreateSet(7, 5, 9),
                                                  ImapStoreDataItem.AddFlagsSilent(ImapMessageFlag.Deleted),
                                                  320162338UL,
                                                  out messages,
                                                  out failedMessageSet));
  
        AssertEqualCaseInsensitive("0004 STORE 7,5,9 (UNCHANGEDSINCE 320162338) +FLAGS.SILENT (\\Deleted)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(7, messages[0].Sequence);
        Assert.AreEqual(320162342UL, messages[0].ModSeq);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Deleted));

        Assert.AreEqual(5, messages[1].Sequence);
        Assert.AreEqual(320162350UL, messages[1].ModSeq);
        Assert.AreEqual(0, messages[1].Flags.Count);

        Assert.AreEqual(9, messages[2].Sequence);
        Assert.AreEqual(320162349UL, messages[2].ModSeq);
        Assert.AreEqual(1, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Has(ImapMessageFlag.Answered));

        var failed = failedMessageSet.ToArray();

        Assert.AreEqual(2, failed.Length);
        Assert.AreEqual(7, failed[0]);
        Assert.AreEqual(9, failed[1]);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestStoreUnchangedSinceWithTaggedNoResponse()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // STORE
        server.EnqueueResponse("* 1 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                               "* 3 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                               "0004 NO [MODIFIED 2] Some of the messages no longer exist.\r\n");

        ImapMessage[] messages;
        ImapSequenceSet failedMessageSet;

        Assert.IsFalse((bool)session.StoreUnchangedSince(ImapSequenceSet.CreateRangeSet(1, 7),
                                                   ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen),
                                                   320172338UL,
                                                   out messages,
                                                   out failedMessageSet));

        AssertEqualCaseInsensitive("0004 STORE 1:7 (UNCHANGEDSINCE 320172338) +FLAGS (\\SEEN)\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(2, messages.Length);

        Assert.AreEqual(1, messages[0].Sequence);
        Assert.AreEqual(320172342UL, messages[0].ModSeq);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Has(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messages[1].Sequence);
        Assert.AreEqual(320172342UL, messages[1].ModSeq);
        Assert.AreEqual(1, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Has(ImapMessageFlag.Seen));

        var failed = failedMessageSet.ToArray();

        Assert.AreEqual(1, failed.Length);
        Assert.AreEqual(2, failed[0]);

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestStoreUnchangedSinceIncapable()
    {
      using (var session = SelectMailbox()) {
        // STORE
        ImapMessage[] messages;
        ImapSequenceSet failedMessageSet;
        session.StoreUnchangedSince(ImapSequenceSet.CreateRangeSet(1, 7),
                                    ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen),
                                    320172338UL,
                                    out messages,
                                    out failedMessageSet);
      }
    }

    [Test]
    public void TestSearch()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");
  
        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Unseen, out matched));
  
        AssertEqualCaseInsensitive("0004 search UNSEEN\r\n",
                                   server.DequeueRequest());
  
        Assert.IsFalse(matched.IsUidSet);

        var messages = matched.ToArray();

        Assert.AreEqual(3, messages.Length);
        Assert.AreEqual(2,    messages[0]);
        Assert.AreEqual(84,   messages[1]);
        Assert.AreEqual(882,  messages[2]);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSearchNotMatched()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("+ OK\r\n");
        server.EnqueueResponse("* SEARCH\r\n" +
                               "0004 OK SEARCH completed\r\n");
  
        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Text("string not in mailbox"), Encoding.UTF8, out matched));
  
        AssertEqualCaseInsensitive("0004 search charset utf-8 TEXT {21}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("string not in mailbox\r\n",
                                   server.DequeueRequest());
  
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);
  
        CloseMailbox(session);
      }
    }

    [Test, Ignore("not implemented")]
    public void TestSearchBadCharset()
    {
      using (var session = SelectMailbox()) {
        // SEARCH transaction
        server.EnqueueResponse("0004 NO [BADCHARSET ONDUL EUC-JP] Unknown charset\r\n");
        server.EnqueueResponse("+ OK\r\n");
        server.EnqueueResponse("0005 OK SEARCH completed\r\n");
  
        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Body("ほげほげ"), out matched));
  
        AssertEqualCaseInsensitive("0004 search charset utf-8 BODY {12}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0005 search charset EUC-JP BODY {8}\r\n",
                                   server.DequeueRequest());
        server.DequeueRequest();
  
        CloseMailbox(session, "0006");
      }
    }

    [Test]
    public void TestSearchModSeq()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 5 6 (MODSEQ 917162500)\r\n" +
                               "0004 OK SEARCH completed\r\n");
  
        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.ModSeq(620162338UL), out matched));

        AssertEqualCaseInsensitive("0004 SEARCH MODSEQ 620162338\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        var messages = matched.ToArray();

        Assert.AreEqual(3, messages.Length);
        Assert.AreEqual(2, messages[0]);
        Assert.AreEqual(5, messages[1]);
        Assert.AreEqual(6, messages[2]);

        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSearchModSeqWithEntry()
    {
      using (var session = SelectMailbox("CONDSTORE")) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 5 6 7 11 12 18 19 20 23 (MODSEQ 917162500)\r\n" +
                               "0004 OK SEARCH completed\r\n");
  
        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft), out matched));

        AssertEqualCaseInsensitive("0004 SEARCH MODSEQ \"/flags/\\\\draft\" all 620162338\r\n",
                                   server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        var messages = matched.ToArray();

        Assert.AreEqual(10, messages.Length);
        Assert.AreEqual(2, messages[0]);
        Assert.AreEqual(5, messages[1]);
        Assert.AreEqual(6, messages[2]);
        Assert.AreEqual(23, messages[9]);

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestSearchModSeqIncapable()
    {
      using (var session = SelectMailbox()) {
        ImapMatchedSequenceSet matched;

        session.Search(ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft), out matched);
      }
    }

    [Test]
    public void TestSearchResultOption()
    {
      using (var session = SelectMailbox("ESEARCH")) {
        // SEARCH transaction
        server.EnqueueResponse("* ESEARCH (TAG \"0004\") MIN 2 COUNT 3\r\n" +
                               "0004 OK SEARCH completed\r\n");
  
        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Min + ImapSearchResultOptions.Count, out matched));
  
        AssertEqualCaseInsensitive("0004 SEARCH RETURN (MIN COUNT) FLAGGED SINCE 1-Feb-1994 NOT (FROM \"Smith\")\r\n",
                                   server.DequeueRequest());
  
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);
        Assert.IsNotNull(matched.Tag);
        Assert.AreEqual("0004", matched.Tag);
        Assert.IsNotNull(matched.Min);
        Assert.AreEqual(2, matched.Min);
        Assert.IsNotNull(matched.Count);
        Assert.AreEqual(3, matched.Count);
  
        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestSearchESearchIncapable()
    {
      using (var session = SelectMailbox()) {
        ImapMatchedSequenceSet matched;

        session.ESearch(ImapSearchCriteria.All, ImapSearchResultOptions.Min, out matched);
      }
    }

    [Test]
    public void TestSearchAndFetchLastResponse()
    {
      using (var session = SelectMailbox("ESEARCH", "SEARCHRES")) {
        // SEARCH transaction
        server.EnqueueResponse("0004 OK SEARCH completed, result saved\r\n");
  
        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Save, out matched));
  
        AssertEqualCaseInsensitive("0004 SEARCH RETURN (SAVE) FLAGGED SINCE 1-Feb-1994 NOT (FROM \"Smith\")\r\n",
                                   server.DequeueRequest());
  
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        // FETCH transaction
        server.EnqueueResponse("* 2 FETCH (UID 14)\r\n" +
                               "* 84 FETCH (UID 100)\r\n" +
                               "* 882 FETCH (UID 1115)\r\n" +
                               "0005 OK completed\r\n");

        ImapMessage[] fetched;
  
        Assert.IsTrue((bool)session.Fetch(matched, ImapFetchDataItem.Uid, out fetched));

        AssertEqualCaseInsensitive("0005 FETCH $ (UID)\r\n",
                                   server.DequeueRequest());

        CloseMailbox(session, "0006");
      }
    }

    [Test]
    public void TestSearchAndSearchAgainLastResponse()
    {
      using (var session = SelectMailbox("ESEARCH", "SEARCHRES")) {
        // SEARCH transaction
        server.EnqueueResponse("0004 OK SEARCH completed, result saved\r\n");
  
        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Save, out matched));
  
        AssertEqualCaseInsensitive("0004 SEARCH RETURN (SAVE) FLAGGED SINCE 1-Feb-1994 NOT (FROM \"Smith\")\r\n",
                                   server.DequeueRequest());
  
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 17 900 901\r\n" +
                               "0005 OK completed\r\n");

        ImapMatchedSequenceSet matched2;
  
        Assert.IsTrue((bool)session.UidSearch(ImapSearchCriteria.Uid(matched) & ImapSearchCriteria.Smaller(4096), out matched2));

        AssertEqualCaseInsensitive("0005 UID SEARCH UID $ SMALLER 4096\r\n",
                                   server.DequeueRequest());

        Assert.IsTrue(matched2.IsUidSet);
        Assert.IsFalse(matched2.IsEmpty);

        var arr = matched2.ToArray();

        Assert.AreEqual(3, arr.Length);
        Assert.AreEqual(17, arr[0]);
        Assert.AreEqual(900, arr[1]);
        Assert.AreEqual(901, arr[2]);

        CloseMailbox(session, "0006");
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestSearchLastResponseIncapable()
    {
      using (var session = SelectMailbox("ESEARCH")) {
        // SEARCH transaction
        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;
  
        session.ESearch(criteria, ImapSearchResultOptions.Save, out matched);
      }
    }

    [Test]
    public void TestSort()
    {
      using (var session = SelectMailbox("SORT")) {
        // SORT transaction
        server.EnqueueResponse("* SORT 5 3 4 1 2\r\n" +
                               "0004 OK SORT completed\r\n");
  
        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date, ImapSearchCriteria.All, out matched));
  
        AssertEqualCaseInsensitive("0004 sort (REVERSE SIZE DATE) utf-8 ALL\r\n",
                                   server.DequeueRequest());
  
        Assert.IsFalse(matched.IsUidSet);

        var messages = matched.ToArray();

        Assert.AreEqual(5, messages.Length);
        Assert.AreEqual(5, messages[0]);
        Assert.AreEqual(3, messages[1]);
        Assert.AreEqual(4, messages[2]);
        Assert.AreEqual(1, messages[3]);
        Assert.AreEqual(2, messages[4]);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestSortResultOptions()
    {
      using (var session = SelectMailbox("SORT", "ESORT")) {
        // SORT transaction
        server.EnqueueResponse("* ESEARCH (TAG \"0004\") UID ALL 23765,23764,23763,23761\r\n" +
                               "0004 OK SORT completed\r\n");
  
        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.UidESort(ImapSortCriteria.DateReverse,
                                       ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                                       ImapSearchResultOptions.All,
                                       out matched));
  
        AssertEqualCaseInsensitive("0004 UID SORT RETURN (ALL) (REVERSE DATE) UTF-8 UNDELETED UNKEYWORD $Junk\r\n",
                                   server.DequeueRequest());
  
        Assert.IsTrue(matched.IsUidSet);
        Assert.AreEqual("0004", matched.Tag);

        var messages = matched.ToArray();

        Assert.AreEqual(4, messages.Length);
        Assert.AreEqual(23765, messages[0]);
        Assert.AreEqual(23764, messages[1]);
        Assert.AreEqual(23763, messages[2]);
        Assert.AreEqual(23761, messages[3]);
  
        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestSortESortIncapable()
    {
      using (var session = SelectMailbox("SORT")) {
        // SORT transaction
        ImapMatchedSequenceSet matched;
  
        Assert.IsTrue((bool)session.UidESort(ImapSortCriteria.DateReverse,
                                       ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                                       ImapSearchResultOptions.All,
                                       out matched));
      }
    }

    [Test]
    public void TestThread()
    {
      using (var session = SelectMailbox("THREAD=REFERENCES")) {
        // THREAD transaction
        server.EnqueueResponse("* THREAD (166)(167)(168)(169)(172)(170)(171)(173)(174 (175)(176)(178)(181)(180))(179)(177 (183)(182)(188)(184)(185)(186)(187)(189))(190)(191)(192)(193)(194 195)(196 (197)(198))(199)(200 202)(201)(203)(204)(205)(206 207)(208)\r\n" +
                               "0004 OK THREAD completed\r\n");
  
        ImapThreadList threadList;
  
        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References, ImapSearchCriteria.All, out threadList));
  
        AssertEqualCaseInsensitive("0004 thread REFERENCES utf-8 ALL\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(25, threadList.Children.Length);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestExpunge()
    {
      using (var session = SelectMailbox()) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 4 EXPUNGE\r\n" +
                               "* 7 EXPUNGE\r\n" +
                               "* 11 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");
  
        long[] expunged;
  
        Assert.IsTrue((bool)session.Expunge(out expunged));
  
        AssertEqualCaseInsensitive("0004 expunge\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(4, expunged.Length);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(4,  expunged[1]);
        Assert.AreEqual(7,  expunged[2]);
        Assert.AreEqual(11, expunged[3]);
  
        CloseMailbox(session);
      }
    }


    [Test]
    public void TestUidExpunge()
    {
      using (var session = SelectMailbox("UIDPLUS")) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");
  
        long[] expunged;
  
        Assert.IsTrue((bool)session.UidExpunge(ImapSequenceSet.CreateUidFromSet(1), out expunged));
  
        AssertEqualCaseInsensitive("0004 UID EXPUNGE 1:*\r\n",
                                   server.DequeueRequest());

        Assert.AreEqual(3, expunged.Length);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(3,  expunged[1]);
        Assert.AreEqual(3,  expunged[2]);
  
        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestExpungeSequenceSet()
    {
      using (var session = SelectMailbox("UIDPLUS")) {
        long[] expunged;

        session.UidExpunge(ImapSequenceSet.CreateFromSet(1), out expunged);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestUidExpungeEmptyUidSet()
    {
      using (var session = SelectMailbox("UIDPLUS")) {
        long[] expunged;

        session.UidExpunge(ImapSequenceSet.CreateUidSet(new long[] {}), out expunged);
      }
    }

    [Test]
    public void TestCopy()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 OK COPY completed\r\n");
  
        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX"));
  
        AssertEqualCaseInsensitive("0004 copy 1:10 \"INBOX\"\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestCopyEmptySequenceSet()
    {
      using (var session = SelectMailbox()) {
        session.Copy(ImapSequenceSet.CreateSet(new long[] {}), "INBOX");
      }
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestCopyEmptyUidSet()
    {
      using (var session = SelectMailbox()) {
        session.Copy(ImapSequenceSet.CreateUidSet(new long[] {}), "INBOX");
      }
    }

    [Test]
    public void TestCopyTryCreate()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
        server.EnqueueResponse("0005 OK CREATE completed\r\n");
        server.EnqueueResponse("0006 OK COPY completed\r\n");
  
        ImapMailbox created;
  
        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateUidRangeSet(1, 10), "INBOX.copyto", out created));
  
        Assert.AreEqual("INBOX.copyto", created.Name);
  
        AssertEqualCaseInsensitive("0004 uid copy 1:10 \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0005 create \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0006 uid copy 1:10 \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session, "0007");
      }
    }

    [Test]
    public void TestCopyDontTryCreate()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
  
        Assert.IsFalse((bool)session.Copy(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX.copyto"));
  
        AssertEqualCaseInsensitive("0004 copy 1:10 \"INBOX.copyto\"\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestCopyWithAppendUidResponseCode()
    {
      using (var session = SelectMailbox()) {
        // COPY transaction
        server.EnqueueResponse("0004 OK [COPYUID 38505 304,319:320 3956:3958] Done\r\n");
  
        ImapCopiedUidSet copied;

        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateRangeSet(2, 4), "meeting", out copied));
  
        AssertEqualCaseInsensitive("0004 COPY 2:4 \"meeting\"\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(copied);
        Assert.AreEqual(38505, copied.UidValidity);

        var copiedUids = copied.CopiedUidSet.ToArray();

        Assert.AreEqual(3, copiedUids.Length);
        Assert.AreEqual(304, copiedUids[0]);
        Assert.AreEqual(319, copiedUids[1]);
        Assert.AreEqual(320, copiedUids[2]);

        var assignedUids = copied.AssignedUidSet.ToArray();

        Assert.AreEqual(3, assignedUids.Length);
        Assert.AreEqual(3956, assignedUids[0]);
        Assert.AreEqual(3957, assignedUids[1]);
        Assert.AreEqual(3958, assignedUids[2]);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppend()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen));
  
        Assert.IsTrue((bool)session.Append(message, "INBOX"));
  
        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("MIME-Version: 1.0\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("test message\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    [Ignore("broken")]
    public void TestAppendNonSynchronizedLiterals()
    {
      using (var session = SelectMailbox("LITERAL+")) {
        // APPEND transaction
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen));
  
        Assert.IsTrue((bool)session.Append(message, "INBOX"));
  
        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33+}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("MIME-Version: 1.0\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("test message\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendTryCreate()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n");
        server.EnqueueResponse("0005 OK CREATE completed\r\n");
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0006 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(new[] {"$label2"}, ImapMessageFlag.Draft));
  
        ImapMailbox created;
  
        Assert.IsTrue((bool)session.Append(message, "INBOX.appendto", out created));
  
        Assert.AreEqual("INBOX.appendto", created.Name);
  
        AssertEqualCaseInsensitive("0004 append \"INBOX.appendto\" (\\Draft $label2) {33}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0005 create \"INBOX.appendto\"\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("0006 append \"INBOX.appendto\" (\\Draft $label2) {33}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("MIME-Version: 1.0\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("test message\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session, "0007");
      }
    }

    [Test]
    public void TestAppendDontTryCreate()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new DateTimeOffset(2008, 2, 25, 15, 1, 12, new TimeSpan(+9, 0, 0)));
  
        var internalDate = "\"25-Feb-2008 15:01:12 +0900\"";
  
        Assert.IsFalse((bool)session.Append(message, "INBOX.appendto"));
  
        AssertEqualCaseInsensitive("0004 append \"INBOX.appendto\" " + internalDate + " {33}\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendRemoveNonApplicableFlags()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen,
                                                              ImapMessageFlag.Recent,
                                                              ImapMessageFlag.AllowedCreateKeywords));
  
        Assert.IsTrue((bool)session.Append(message, "INBOX"));
  
        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("MIME-Version: 1.0\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("test message\r\n",
                                   server.DequeueRequest());
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestAppendWithAppendUidResponseCode()
    {
      using (var session = SelectMailbox()) {
        // APPEND transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0004 OK [APPENDUID 38505 3955] APPEND completed\r\n");
  
        var message = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                      new ImapMessageFlagList(ImapMessageFlag.Seen));
  
        ImapAppendedUidSet appended;

        Assert.IsTrue((bool)session.Append(message, "INBOX", out appended));

        AssertEqualCaseInsensitive("0004 append \"INBOX\" (\\Seen) {33}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("MIME-Version: 1.0\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("test message\r\n",
                                   server.DequeueRequest());

        Assert.IsNotNull(appended);
        Assert.AreEqual(38505, appended.UidValidity);
        Assert.AreEqual("3955", appended.ToString());
        CloseMailbox(session);
      }
    }

    [Test]
    [Ignore("broken")]
    public void TestAppendMultiple()
    {
      using (var session = SelectMailbox("MULTIAPPEND")) {
        // APPEND transaction
        server.EnqueueResponse("+ Ready for literal data\r\n");
        server.EnqueueResponse("+ Ready for literal data\r\n");
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message1 = new ImapMessage(Encoding.ASCII.GetBytes("Date: Mon, 7 Feb 1994 21:52:25 -0800 (PST)\r\n" +
                                       "From: Fred Foobar <foobar@Blurdybloop.example.COM>\r\n" +
                                       "Subject: afternoon meeting\r\n" +
                                       "To: mooch@owatagu.example.net\r\n" +
                                       "Message-Id: <B27397-0100000@Blurdybloop.example.COM>\r\n" +
                                       "MIME-Version: 1.0\r\n" +
                                       "Content-Type: TEXT/PLAIN; CHARSET=US-ASCII\r\n" +
                                       "\r\n" +
                                       "Hello Joe, do you think we can meet at 3:30 tomorrow?\r\n"),
                                       new ImapMessageFlagList(ImapMessageFlag.Seen));
        var message2 = new ImapMessage(Encoding.ASCII.GetBytes("Date: Mon, 7 Feb 1994 22:43:04 -0800 (PST)\r\n" +
                                       "From: Joe Mooch <mooch@OWaTaGu.example.net>\r\n" +
                                       "Subject: Re: afternoon meeting\r\n" +
                                       "To: foobar@blurdybloop.example.com\r\n" +
                                       "Message-Id: <a0434793874930@OWaTaGu.example.net>\r\n" +
                                       "MIME-Version: 1.0\r\n" +
                                       "Content-Type: TEXT/PLAIN; CHARSET=US-ASCII\r\n" +
                                       "\r\n" +
                                       "3:30 is fine with me.\r\n"),
                                       new DateTimeOffset(new DateTime(1994, 2, 7, 22, 43, 04), new TimeSpan(-8, 0, 0)),
                                       new ImapMessageFlagList(ImapMessageFlag.Seen));

        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages"));

        AssertEqualCaseInsensitive("0004 APPEND \"saved-messages\" (\\Seen) {329}\r\n", server.DequeueRequest());
        Assert.IsTrue(server.DequeueRequest().StartsWith("Date"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("From"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("Subject"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("To"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("Message-Id"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("MIME-Version"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("Content-Type"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("\r\n"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("Hello"));
        AssertEqualCaseInsensitive(" (\\Seen) \"07-Feb-1994 22:43:04 -0800\" {295}\r\n", server.DequeueRequest()); // XXX
        Assert.IsTrue(server.DequeueRequest().StartsWith("Date"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("From"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("Subject"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("To"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("Message-Id"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("MIME-Version"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("Content-Type"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("\r\n"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("3:30"));
        Assert.IsTrue(server.DequeueRequest().StartsWith("\r\n"));

        CloseMailbox(session);
      }
    }

    [Test]
    [Ignore("broken")]
    public void TestAppendMultipleWithNonSyncedLiteral()
    {
      using (var session = SelectMailbox("MULTIAPPEND", "LITERAL+")) {
        // APPEND transaction
        server.EnqueueResponse("0004 OK APPEND completed\r\n");
  
        var message1 = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message1"));
        var message2 = new ImapMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message2"));
        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages"));

        AssertEqualCaseInsensitive("0004 APPEND \"saved-messages\" {34+}\r\n", server.DequeueRequest());
        AssertEqualCaseInsensitive("MIME-Version: 1.0\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("test message1 {34+}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("MIME-Version: 1.0\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("test message2\r\n",
                                   server.DequeueRequest());

        CloseMailbox(session);
      }
    }

    [Test]
    [ExpectedException(typeof(ImapIncapableException))]
    public void TestAppendMultipleIncapable()
    {
      using (var session = SelectMailbox()) {
        session.HandlesIncapableAsException = true;
        // APPEND transaction
        session.AppendMultiple(new[] {
          new ImapMessage(Encoding.ASCII.GetBytes("message")),
          new ImapMessage(Encoding.ASCII.GetBytes("message2")),
        }, "saved-messages");
      }
    }

    [Test]
    public void TestUnselect()
    {
      using (var session = SelectMailbox("UNSELECT")) {
        // EXPUNGE transaction
        server.EnqueueResponse("0004 OK UNSELECT completed\r\n");
  
        Assert.IsTrue((bool)session.Unselect());
  
        AssertEqualCaseInsensitive("0004 unselect\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      }
    }

    private ImapSession Authenticate(params string[] additionalCapabilities)
    {
      // greeting and login transaction
      server.EnqueueResponse("* OK ImapSimulatedServer ready\r\n");

      var session = new ImapSession(host, port);

      session.Credentials = credential;

      server.EnqueueResponse("0000 OK authenticated\r\n");

      session.Login();

      Assert.AreEqual(new Uri(string.Format("imap://{0}@{1}:{2}/", username, host, port)), session.Authority);

      server.DequeueRequest();

      // CAPABILITY transaction
      server.EnqueueResponse(string.Format("* CAPABILITY {0} IMAP4rev1\r\n", string.Join(" ", additionalCapabilities)) +
                             "0001 OK CAPABILITY completed\r\n");

      session.Capability();

      server.DequeueRequest();

      return session;
    }

    [Test]
    public void TestCreate()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.List("", out mailboxes);
  
        server.DequeueRequest();
  
        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
  
        // CREATE transaction
        server.EnqueueResponse("0003 OK CREATE completed\r\n");
  
        ImapMailbox createdMailbox;
  
        Assert.IsTrue((bool)session.Create("INBOX.日本語", out createdMailbox));
  
        AssertEqualCaseInsensitive("0003 create \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
  
        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("INBOX.日本語", createdMailbox.Name);
      }
    }

    [Test]
    public void TestDelete()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.List("", out mailboxes);
  
        server.DequeueRequest();
  
        Assert.AreEqual(2, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.AreEqual("INBOX.日本語", mailboxes[1].Name);
        Assert.AreEqual(".", mailboxes[1].HierarchyDelimiter);
  
        // DELETE transaction
        server.EnqueueResponse("0003 OK DELETE completed\r\n");
  
        Assert.IsTrue((bool)session.Delete(mailboxes[1]));
  
        AssertEqualCaseInsensitive("0003 delete \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    [ExpectedException(typeof(InvalidOperationException))]
    public void TestDeleteSelectedMailbox()
    {
      using (var session = SelectMailbox()) {
        // DELETE transaction
        session.Delete("INBOX");
      }
    }

    [Test]
    public void TestRename()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.List("", out mailboxes);
  
        server.DequeueRequest();
  
        Assert.AreEqual(2, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.AreEqual("INBOX.日本語", mailboxes[1].Name);
        Assert.AreEqual(".", mailboxes[1].HierarchyDelimiter);
  
        // REMANE transaction
        server.EnqueueResponse("0003 OK RENAME completed\r\n");
  
        ImapMailbox renamedMailbox;
  
        Assert.IsTrue((bool)session.Rename(mailboxes[1], "INBOX.☺", out renamedMailbox));
  
        AssertEqualCaseInsensitive("0003 rename \"INBOX.&ZeVnLIqe-\" \"INBOX.&Jjo-\"\r\n",
                                   server.DequeueRequest());
  
        Assert.IsNotNull(renamedMailbox);
        Assert.AreEqual("INBOX.☺", renamedMailbox.Name);
      }
    }

    [Test]
    [ExpectedException(typeof(InvalidOperationException))]
    public void TestRenameSelectedMailbox()
    {
      using (var session = SelectMailbox()) {
        // REMANE transaction
        session.Rename("INBOX", "INBOX2");
      }
    }

    [Test]
    public void TestRenameMailboxReferralAsException()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.List("", out mailboxes);
  
        server.DequeueRequest();
  
        // REMANE transaction
        server.EnqueueResponse("0003 NO [REFERRAL IMAP://user;AUTH=*@SERVER1/FOO IMAP://user;AUTH=*@SERVER2/BAR] Unable to rename mailbox across servers\r\n");
  
        try {
          session.HandlesReferralAsException = true;
  
          ImapMailbox renamedMailbox;
          session.Rename(mailboxes[1], "INBOX.☺", out renamedMailbox);
  
          Assert.Fail("renamed without exception");
        }
        catch (ImapMailboxReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Unable to rename mailbox across servers"));
          Assert.AreEqual(2, ex.Referrals.Length);
          Assert.AreEqual(new Uri("IMAP://user;AUTH=*@SERVER1/FOO"), ex.Referrals[0]);
          Assert.AreEqual(new Uri("IMAP://user;AUTH=*@SERVER2/BAR"), ex.Referrals[1]);
        }
      }
    }

    [Test]
    public void TestSubscribe()
    {
      using (var session = Authenticate()) {
        // LIST transaction
        server.EnqueueResponse("* LIST () \".\" \"INBOX\"\r\n" +
                               "* LIST () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LIST completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.List("", out mailboxes);
  
        server.DequeueRequest();
  
        // SUBSCRIBE transaction
        server.EnqueueResponse("0003 OK SUBSCRIBE completed\r\n");
  
        Assert.IsTrue((bool)session.Subscribe(mailboxes[1]));
  
        AssertEqualCaseInsensitive("0003 subscribe \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestUnsubscribe()
    {
      using (var session = Authenticate()) {
        // LSUB transaction
        server.EnqueueResponse("* LSUB () \".\" \"INBOX\"\r\n" +
                               "* LSUB () \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                               "0002 OK LSUB completed\r\n");
  
        ImapMailbox[] mailboxes;
  
        session.Lsub("INBOX*", out mailboxes);
  
        server.DequeueRequest();
  
        // UNSUBSCRIBE transaction
        server.EnqueueResponse("0003 OK UNSUBSCRIBE completed\r\n");
  
        Assert.IsTrue((bool)session.Unsubscribe(mailboxes[1]));
  
        AssertEqualCaseInsensitive("0003 unsubscribe \"INBOX.&ZeVnLIqe-\"\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    public void TestNamespace()
    {
      using (var session = Authenticate("NAMESPACE")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Namespace));
  
        // NAMESPACE transaction
        server.EnqueueResponse("* NAMESPACE ((\"\" \"/\")) ((\"~\" \"/\")) ((\"#shared/\" \"/\")(\"#public/\" \"/\")(\"#ftp/\" \"/\")(\"#news.\" \".\"))\r\n" +
                               "0002 OK NAMESPACE command completed\r\n");
  
        ImapNamespace namespaces = null;
  
        Assert.IsTrue((bool)session.Namespace(out namespaces));
  
        AssertEqualCaseInsensitive("0002 NAMESPACE\r\n",
                                   server.DequeueRequest());
  
        Assert.IsNotNull(namespaces);
        Assert.AreEqual(1, namespaces.PersonalNamespaces.Length);
        Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
        Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);
  
        Assert.AreEqual(1, namespaces.OtherUsersNamespaces.Length);
        Assert.AreEqual("~", namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.AreEqual("/", namespaces.OtherUsersNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, namespaces.OtherUsersNamespaces[0].Extensions.Count);
  
        Assert.AreEqual(4, namespaces.SharedNamespaces.Length);
        Assert.AreEqual("#shared/", namespaces.SharedNamespaces[0].Prefix);
        Assert.AreEqual("#public/", namespaces.SharedNamespaces[1].Prefix);
        Assert.AreEqual("#ftp/", namespaces.SharedNamespaces[2].Prefix);
        Assert.AreEqual("#news.", namespaces.SharedNamespaces[3].Prefix);
  
        Assert.IsTrue(namespaces.PersonalNamespaces.Length == session.Namespaces.PersonalNamespaces.Length);
        Assert.IsTrue(namespaces.OtherUsersNamespaces.Length == session.Namespaces.OtherUsersNamespaces.Length);
        Assert.IsTrue(namespaces.SharedNamespaces.Length == session.Namespaces.SharedNamespaces.Length);
        Assert.IsTrue(namespaces.PersonalNamespaces[0].Prefix == session.Namespaces.PersonalNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.OtherUsersNamespaces[0].Prefix == session.Namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[0].Prefix == session.Namespaces.SharedNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[1].Prefix == session.Namespaces.SharedNamespaces[1].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[2].Prefix == session.Namespaces.SharedNamespaces[2].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[3].Prefix == session.Namespaces.SharedNamespaces[3].Prefix);
      }
    }

    [Test]
    public void TestIdle()
    {
      using (var session = SelectMailbox("IDLE")) {
        Assert.IsFalse(session.IsIdling);

        // not idling
        try {
          session.EndIdle(null);
          Assert.Fail("no exception occured");
        }
        catch {
        }

        session.StatusUpdated += delegate(object sender, ImapStatusUpdatedEventArgs e) {
          Assert.IsFalse(e.UpdatedStatus.MessageAttributes.Length == 0 &&
                         e.UpdatedStatus.Expunged.Length == 0 &&
                         e.UpdatedStatus.Exists == null &&
                         e.UpdatedStatus.Recent == null &&
                         e.UpdatedStatus.Flags == null,
                         "status changed during idling");
        };
  
        // IDLE
        server.EnqueueResponse("* 2 EXPUNGE\r\n" +
                               "+ idling\r\n" +
                               "* 3 EXISTS\r\n");
  
        var asyncResult = session.BeginIdle();

        Assert.IsNotNull(asyncResult);

        Thread.Sleep(250);

        Assert.IsTrue((bool)session.IsIdling);
  
        AssertEqualCaseInsensitive("0004 idle\r\n",
                                   server.DequeueRequest());
  
        Thread.Sleep(1000);
  
        Assert.IsFalse(3 == session.SelectedMailbox.ExistsMessage);
  
        server.EnqueueResponse("* 4 EXISTS\r\n" +
                               "0004 OK IDLE terminated\r\n");
  
        Assert.IsTrue((bool)session.EndIdle(asyncResult));
  
        AssertEqualCaseInsensitive("done\r\n",
                                   server.DequeueRequest());
  
        Assert.IsFalse(session.IsIdling);
        Assert.AreEqual(4, session.SelectedMailbox.ExistsMessage);
  
        Assert.AreEqual(ImapUpdatedStatusKind.Expunged | ImapUpdatedStatusKind.Exists,
                        session.LastUpdatedStatus.UpdatedKind);

        Assert.AreEqual(0, session.LastUpdatedStatus.MessageAttributes.Length);
        Assert.IsNotNull(session.LastUpdatedStatus.Expunged);
        Assert.AreEqual(1, session.LastUpdatedStatus.Expunged.Length);
        Assert.AreEqual(2, session.LastUpdatedStatus.Expunged[0]);
        Assert.IsNotNull(session.LastUpdatedStatus.Exists);
        Assert.AreEqual(4, (long)session.LastUpdatedStatus.Exists);
        Assert.IsNull(session.LastUpdatedStatus.Recent);
  
        CloseMailbox(session);
      }
    }

    [Test]
    public void TestEnable()
    {
      using (var session = Authenticate("ENABLE", "X-GOOD-IDEA")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.Enable));
  
        // ENABLE transaction
        server.EnqueueResponse("* ENABLED X-GOOD-IDEA\r\n" +
                               "0002 OK ENABLE completed\r\n");
  
        ImapCapabilityList enabledCapas;
  
        Assert.IsTrue((bool)session.Enable(out enabledCapas, "CONDSTORE", "X-GOOD-IDEA"));
  
        AssertEqualCaseInsensitive("0002 ENABLE CONDSTORE X-GOOD-IDEA\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, enabledCapas.Count);
        Assert.IsTrue(enabledCapas.Has("X-GOOD-IDEA"));
      }
    }

    [Test]
    public void TestGenericCommandNoWithArguments()
    {
      using (var session = Authenticate()) {
        // generic command transaction
        server.EnqueueResponse("* X-EXT-COMMAND data\r\n" +
                               "0002 OK X-EXT-COMMAND completed\r\n");
  
        ImapDataResponse[] dataResps;
  
        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps));
  
        AssertEqualCaseInsensitive("0002 X-EXT-COMMAND\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(1, dataResps.Length);
        Assert.AreEqual(2, dataResps[0].Data.Length);
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[0].Format);
        Assert.AreEqual("X-EXT-COMMAND", dataResps[0].Data[0].Text.ToString());
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[1].Format);
        Assert.AreEqual("data", dataResps[0].Data[1].Text.ToString());
      }
    }

    [Test]
    public void TestGenericCommandWithArguments()
    {
      using (var session = Authenticate()) {
        // generic command transaction
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse("0002 OK X-EXT-COMMAND completed\r\n");
  
        ImapDataResponse[] dataResps;
  
        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps,
                                           new ImapParenthesizedString("arg1",
                                                                       "arg2",
                                                                       new ImapParenthesizedString("arg3-1", "arg3-2"),
                                                                       new ImapLiteralString("ascii text", ImapEncoding.Transfer8Bit)),
                                           new ImapParenthesizedString(new ImapQuotedString("quoted"),
                                                                       "nonquoted")
                                           ));
  
        AssertEqualCaseInsensitive("0002 X-EXT-COMMAND (arg1 arg2 (arg3-1 arg3-2) {10}\r\n",
                                   server.DequeueRequest());
        AssertEqualCaseInsensitive("ascii text) (\"quoted\" nonquoted)\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(0, dataResps.Length);
      }
    }

    [Test]
    public void TestId()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        // ID transaction
        server.EnqueueResponse("* ID NIL\r\n" +
                               "0002 OK ID completed\r\n");
  
        Dictionary<string, string> serverParameterList;
  
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"name", "sodr"},
          {"version", "19.34"},
          {"vendor", "Pink Floyd Music Limited"},
        }, out serverParameterList));
  
        AssertEqualCaseInsensitive("0002 ID (\"name\" \"sodr\" \"version\" \"19.34\" \"vendor\" \"Pink Floyd Music Limited\")\r\n",
                                   server.DequeueRequest());
  
        Assert.AreEqual(0, serverParameterList.Count);
      }
    }

    [Test]
    public void TestIdParameterNil()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
    
        // ID transaction
        server.EnqueueResponse("* ID NIL\r\n" +
                               "0002 OK ID completed\r\n");
    
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"name", null},
          {"version", null},
          {"vendor", null},
        }));
    
        AssertEqualCaseInsensitive("0002 ID (\"name\" NIL \"version\" NIL \"vendor\" NIL)\r\n",
                                   server.DequeueRequest());
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIdTooMatchFieldValuePair()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        for (var i = 0; i < 31; i++) {
          paramList.Add(string.Format("field{0}", i), string.Format("value{0}", i));
        }
  
        session.ID(paramList);
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIdTooLongField()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {new string('x', 31), "value"},
        }));
  
        session.ID(paramList);
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIdTooLongValue()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"field", new string('x', 1025)},
        }));
  
        session.ID(paramList);
      }
    }

    [Test]
    [ExpectedException(typeof(ArgumentException))]
    public void TestIdContaisNoAsciiChar()
    {
      using (var session = Authenticate("ID")) {
        Assert.IsTrue(session.ServerCapabilities.Has(ImapCapability.ID));
  
        var paramList = new Dictionary<string, string>();
  
        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"オワタ", "＼(＾o＾)／"},
        }));
  
        session.ID(paramList);
      }
    }

    private int port;
    private string host;
    private Uri uri;
    private const string username = "imapuser";
    private const string password = "mikumiku";
    private NetworkCredential credential;

    private Encoding latin1Encoding = Encoding.GetEncoding("latin1");

    private ImapSimulatedServer server;
  }
}
