using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapSearchCriteriaTest {
    [Test]
    public void TestDate()
    {
      var criteria = ImapSearchCriteria.Since(new DateTime(2008, 2, 25));

      Assert.AreEqual("SINCE 25-Feb-2008", criteria.ToString());
    }

    [Test]
    public void TestSize()
    {
      var criteria = ImapSearchCriteria.Larger(72);

      Assert.AreEqual("LARGER 72", criteria.ToString());
    }

    [Test]
    public void TestUid()
    {
      var criteria = ImapSearchCriteria.Uid(ImapSequenceSet.CreateSet(1, 4, 5));

      Assert.AreEqual("UID 1,4,5", criteria.ToString());
    }

    [Test]
    public void TestAnd1()
    {
      var criteria1 = ImapSearchCriteria.From("from");
      var criteria2 = ImapSearchCriteria.Recent;
      var expected = "FROM \"from\" RECENT";

      Assert.AreEqual(expected, (criteria1 & criteria2).ToString());
      Assert.AreEqual(expected, criteria1.And(criteria2).ToString());
      Assert.AreEqual(expected, ImapSearchCriteria.And(criteria1, criteria2).ToString());
    }

    [Test]
    public void TestAnd2()
    {
      var criteria1 = ImapSearchCriteria.To("to");
      var criteria2 = ImapSearchCriteria.Keyword("label1");

      Assert.AreEqual("TO \"to\" KEYWORD label1", (criteria1 & criteria2).ToString());
      Assert.AreEqual("KEYWORD label1 TO \"to\"", (criteria2 & criteria1).ToString());
    }

    [Test]
    public void TestAnd3()
    {
      var criteria = ImapSearchCriteria.Flagged & ImapSearchCriteria.Undeleted;

      Assert.AreEqual("FLAGGED UNDELETED", criteria.ToString());
    }

    [Test]
    public void TestOr1()
    {
      var criteria1 = ImapSearchCriteria.Recent;
      var criteria2 = ImapSearchCriteria.Keyword("label1");
      var expected = "OR (RECENT) (KEYWORD label1)";

      Assert.AreEqual(expected, (criteria1 | criteria2).ToString());
      Assert.AreEqual(expected, criteria1.Or(criteria2).ToString());
      Assert.AreEqual(expected, ImapSearchCriteria.Or(criteria1, criteria2).ToString());
    }

    [Test]
    public void TestOr2()
    {
      var criteria1 = ImapSearchCriteria.From("from") & ImapSearchCriteria.Keyword("label1");
      var criteria2 = ImapSearchCriteria.Recent;

      Assert.AreEqual("OR (FROM \"from\" KEYWORD label1) (RECENT)", (criteria1 | criteria2).ToString());
      Assert.AreEqual("OR (RECENT) (FROM \"from\" KEYWORD label1)", (criteria2 | criteria1).ToString());
    }

    [Test]
    public void TestNot1()
    {
      var criteria = ImapSearchCriteria.Recent;
      var expected = "NOT (RECENT)";

      Assert.AreEqual(expected, (!criteria).ToString());
      Assert.AreEqual(expected, criteria.Not().ToString());
      Assert.AreEqual(expected, ImapSearchCriteria.Not(criteria).ToString());
    }

    [Test]
    public void TestAndOrNot1()
    {
      var criteria1 = ImapSearchCriteria.From("from") & ImapSearchCriteria.Keyword("label1");
      var criteria2 = ImapSearchCriteria.Recent;
      var criteria = !(criteria1 | criteria2);

      Assert.AreEqual("NOT (OR (FROM \"from\" KEYWORD label1) (RECENT))", criteria.ToString());
    }

    [Test]
    public void TestAndOrNot2()
    {
      var criteria1 = ImapSearchCriteria.From("from") | ImapSearchCriteria.Keyword("label1");
      var criteria2 = ImapSearchCriteria.Recent;
      var criteria = criteria1 & !criteria2;

      Assert.AreEqual("OR (FROM \"from\") (KEYWORD label1) NOT (RECENT)", criteria.ToString());
    }

    [Test]
    public void TestCombineWithContainsRequiredCapability()
    {
      var criteria1 = ImapSearchCriteria.ModSeq(720162338UL);
      var criteria2 = ImapSearchCriteria.Larger(50000);

      var criteria = !criteria1 | criteria2;

      Assert.AreEqual("OR (NOT (MODSEQ 720162338)) (LARGER 50000)", criteria.ToString());

      var caps = (criteria as IImapMultipleExtension).RequiredCapabilities;

      Assert.AreEqual(1, caps.Length);
      Assert.AreEqual(ImapCapability.CondStore, caps[0]);
    }

    [Test]
    public void TestSearchInThreadAndThreadRefs()
    {
      var criteria = ImapSearchCriteria.InThread & ImapSearchCriteria.MessageId("4321.1234321@example.com");

      Assert.AreEqual("INTHREAD MESSAGEID <4321.1234321@example.com>", criteria.ToString());

      var caps = (criteria as IImapMultipleExtension).RequiredCapabilities;

      Assert.AreEqual(2, caps.Length);

      Assert.Contains(ImapCapability.SearchInThread, caps);
      Assert.Contains(ImapCapability.ThreadRefs, caps);
    }

    [Test]
    public void TestFromUri1()
    {
      //   imap://<iserver>/<enc_mailbox>[uidvalidity][?<enc_search>]
      var uri = new Uri("imap://;AUTH=*@minbari.org/gray%20council?SUBJECT%20shadows");
      var criteria = ImapSearchCriteria.FromUri(uri);

      Assert.AreEqual("gray council", ImapStyleUriParser.GetMailbox(uri));
      Assert.AreEqual("?SUBJECT%20shadows", uri.Query);
      Assert.AreEqual(ImapUriForm.SearchMessages, ImapStyleUriParser.GetUriForm(uri));

      Assert.AreEqual("SUBJECT shadows", criteria.ToString());
    }

    [Test]
    public void TestFromUri2()
    {
      var uri = new Uri("imap://john;AUTH=*@minbari.example.org/babylon5/personel?charset%20UTF-8%20SUBJECT%20%7B14+%7D%0D%0A%D0%98%D0%B2%D0%B0%D0%BD%D0%BE%D0%B2%D0%B0");
      var criteria = ImapSearchCriteria.FromUri(uri);

      Assert.AreEqual(ImapUriForm.SearchMessages, ImapStyleUriParser.GetUriForm(uri));
      Assert.AreEqual("charset UTF-8 SUBJECT {14+}\r\nИванова", criteria.ToString());
    }
  }
}
