// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.Threading;
using System.Security.Authentication;

namespace Smdn.Net.Imap4.Protocol {
  public abstract class ImapConnectionBase : IDisposable {
    /*
     * class members
     */
    private static int connectionIdCount = 0;

    /*
     * properties
     */
    public int Id {
      get { return id; }
    }

    public abstract string Host { get; }

    public int Port {
      get
      {
        CheckDisposed();
        CheckConnected();
        return remoteEndPoint.Port;
      }
    }

    public IPEndPoint RemoteEndPoint {
      get
      {
        CheckDisposed();
        CheckConnected();
        return remoteEndPoint;
      }
    }

    public IPEndPoint LocalEndPoint {
      get
      {
        CheckDisposed();
        CheckConnected();
        return localEndPoint;
      }
    }

    public bool IsSslConnection {
      get
      {
        CheckDisposed();
        CheckConnected();

        return (stream.InnerStream is SslStream);
      }
    }

    public bool IsIdling {
      get
      {
        CheckDisposed();
        CheckConnected();
        return isIdling;
      }
      protected set { isIdling = value; }
    }

    protected Socket Socket {
      get
      {
        CheckDisposed();
        CheckConnected();
        return client.Client;
      }
    }

    protected ImapSender Sender {
      get
      {
        CheckDisposed();
        CheckConnected();
        return sender;
      }
    }

    protected ImapReceiver Receiver {
      get
      {
        CheckDisposed();
        CheckConnected();
        return receiver;
      }
    }

    protected ImapConnectionBase()
    {
      this.id = Interlocked.Increment(ref connectionIdCount);

      if (connectionIdCount == int.MaxValue)
        Interlocked.Exchange(ref connectionIdCount, 0);
    }

    ~ImapConnectionBase()
    {
      Dispose(false);
    }

    void IDisposable.Dispose()
    {
      Close();
    }

    public void Close()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    protected virtual void Dispose(bool disposing)
    {
      if (disposed)
        return;

      if (disposing) {
        try {
          if (client != null)
            client.Close();
        }
        finally {
          client = null;
          stream = null;
        }
      }

      disposed = true;
    }

    protected void SetClient(TcpClient client, ImapUpgradeConnectionStreamCallback createAuthenticatedStreamCallback)
    {
      try {
        this.client = client;

        SetStream(client.GetStream());

        if (createAuthenticatedStreamCallback != null)
          UpgradeStream(createAuthenticatedStreamCallback);

        this.remoteEndPoint = client.Client.RemoteEndPoint as IPEndPoint;
        this.localEndPoint  = client.Client.LocalEndPoint  as IPEndPoint;
      }
      catch (SocketException ex) {
        if (this.client != null)
          this.client = null;

        throw new ImapConnectionException(ex.Message, ex);
      }
    }

    private void SetStream(Stream newStream)
    {
      stream    = CreateBufferedStream(newStream);
      receiver  = CreateReceiver(stream);
      sender    = CreateSender(stream);
    }

    protected virtual ImapBufferedStream CreateBufferedStream(Stream stream)
    {
      return new ImapBufferedStream(stream);
    }

    protected abstract ImapSender CreateSender(ImapBufferedStream stream);

    protected abstract ImapReceiver CreateReceiver(ImapBufferedStream stream);

    protected void UpgradeStream(ImapUpgradeConnectionStreamCallback upgradeStreamCallback)
    {
      CheckDisposed();
      CheckConnected();

      try {
        SetStream(upgradeStreamCallback(this, stream.InnerStream));
      }
      catch (AuthenticationException ex) {
        throw new ImapConnectionException(string.Format("upgrading stream failed (callback: {0})", upgradeStreamCallback.Method), ex);
      }
      catch (IOException ex) {
        throw new ImapConnectionException(string.Format("upgrading stream failed (callback: {0})", upgradeStreamCallback.Method), ex);
      }
    }

    public override string ToString()
    {
      return string.Format("{{Host={0}, RemoteEndPoint={1}, LocalEndPoint={2}, Connected={3}}}",
                           Host,
                           remoteEndPoint,
                           localEndPoint,
                           Socket.Connected);
    }

    protected void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    protected void CheckConnected()
    {
      if (client == null)
        throw new InvalidOperationException("not connected");
    }

    private readonly int id = 0;
    private bool disposed = false;
    private IPEndPoint remoteEndPoint;
    private IPEndPoint localEndPoint;
    private bool isIdling = false;
    private TcpClient client = null;
    private ImapBufferedStream stream = null;
    private ImapReceiver receiver = null;
    private ImapSender sender = null;
  }
}
