using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapSequenceSetTest {
    [Test]
    public void TestSet()
    {
      Assert.AreEqual("2", ImapSequenceSet.CreateSet(2).ToString());
      Assert.AreEqual("2,4,5,7", ImapSequenceSet.CreateSet(2, 4, 5, 7).ToString());
      Assert.AreEqual("1,2,3,4,5,6,7,8", ImapSequenceSet.CreateSet(new long[] {1, 2, 3, 4, 5, 6, 7, 8}).ToString());
      Assert.AreEqual("4,5,6,7", ImapSequenceSet.CreateSet(new long[] {1, 2, 3, 4, 5, 6, 7, 8}, 3, 4).ToString());
    }

    [Test]
    public void TestSetToArray()
    {
      var s = ImapSequenceSet.CreateSet(2, 4, 5, 7).ToArray();

      Assert.AreEqual(4, s.Length);
      Assert.AreEqual(2, s[0]);
      Assert.AreEqual(4, s[1]);
      Assert.AreEqual(5, s[2]);
      Assert.AreEqual(7, s[3]);
    }

    [Test]
    public void TestRange()
    {
      Assert.AreEqual("15:18", ImapSequenceSet.CreateRangeSet(15, 18).ToString());
      //Assert.AreEqual("18:15", ImapSequenceSet.CreateRangeSet(18, 15).ToString());
    }

    [Test]
    public void TestRangeToArray()
    {
      var s = ImapSequenceSet.CreateRangeSet(15, 18).ToArray();

      Assert.AreEqual(4, s.Length);
      Assert.AreEqual(15, s[0]);
      Assert.AreEqual(16, s[1]);
      Assert.AreEqual(17, s[2]);
      Assert.AreEqual(18, s[3]);
    }

    [Test]
    public void TestFrom()
    {
      Assert.AreEqual("72:*", ImapSequenceSet.CreateFromSet(72).ToString());
    }

    [Test]
    [ExpectedException(typeof(InvalidOperationException))]
    public void TestFromToArray()
    {
      ImapSequenceSet.CreateFromSet(72).ToArray();
    }

    [Test]
    public void TestTo()
    {
      Assert.AreEqual("*:25", ImapSequenceSet.CreateToSet(25).ToString());
    }

    [Test]
    [ExpectedException(typeof(InvalidOperationException))]
    public void TestToToArray()
    {
      ImapSequenceSet.CreateToSet(25).ToArray();
    }

    [Test]
    public void TestAll()
    {
      Assert.AreEqual("*", ImapSequenceSet.CreateAllSet().ToString());
    }

    [Test]
    [ExpectedException(typeof(InvalidOperationException))]
    public void TestAllToArray()
    {
      ImapSequenceSet.CreateAllSet().ToArray();
    }

    [Test]
    public void TestCombine()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(1, 3);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      Assert.AreEqual("1:3,5,6,8,9", ImapSequenceSet.Combine(set1, set2).ToString());
      Assert.AreEqual("1:3,5,6,8,9", (set1 + set2).ToString());
      Assert.AreEqual("1:3,5,6,8,9", set1.CombineWith(set2).ToString());

      Assert.AreEqual("5,6,8,9,1:3", (set2 + set1).ToString());
    }

    [Test]
    public void TestCombineWithAll()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(1, 3);
      var set2 = ImapSequenceSet.CreateAllSet();

      var combined = set1 + set2;

      Assert.AreEqual("*", combined.ToString());
    }

    [Test]
    public void TestCombinedToArray()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(1, 3);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      var combined = set1 + set2;

      var arr = combined.ToArray();

      Assert.AreEqual(7, arr.Length);
      Assert.AreEqual(1, arr[0]);
      Assert.AreEqual(2, arr[1]);
      Assert.AreEqual(3, arr[2]);
      Assert.AreEqual(5, arr[3]);
      Assert.AreEqual(6, arr[4]);
      Assert.AreEqual(8, arr[5]);
      Assert.AreEqual(9, arr[6]);
    }

    [Test]
    [ExpectedException(typeof(InvalidOperationException))]
    public void TestCombinedWithFromSetToArray()
    {
      var set1 = ImapSequenceSet.CreateFromSet(10);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      (set1 + set2).ToArray();
    }

    [Test]
    public void TestIsEmpty()
    {
      Assert.IsTrue(ImapSequenceSet.CreateSet(new long[] {}).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateFromSet(1).IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidFromSet(1).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateToSet(3).IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidToSet(3).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateRangeSet(1, 3).IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidRangeSet(1, 3).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateAllSet().IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidAllSet().IsEmpty);

      Assert.IsFalse(ImapSequenceSet.Combine(ImapSequenceSet.CreateSet(new long[] {}), ImapSequenceSet.CreateSet(new long[] {1})).IsEmpty);
    }
  }
}