using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapStyleUriParserTests {
    [Test]
    public void TestCheckUriScheme()
    {
      try {
        ImapStyleUriParser.GetUid(null);
        Assert.Fail("ArgumentNullException not thrown");
      }
      catch (ArgumentNullException) {
      }

      try {
        ImapStyleUriParser.GetUid(new Uri("http://localhost/"));
        Assert.Fail("ArgumentException not thrown");
      }
      catch (ArgumentException) {
      }

      try {
        ImapStyleUriParser.GetUid(new Uri("imap://localhost/"));
        ImapStyleUriParser.GetUid(new Uri("imaps://localhost/"));
        ImapStyleUriParser.GetUid(new Uri("IMAP://localhost/"));
        ImapStyleUriParser.GetUid(new Uri("IMAPS://localhost/"));
      }
      catch (ArgumentException) {
        Assert.Fail("ArgumentException thrown");
      }
    }

    [Test]
    public void TestGetAuthority()
    {
      Assert.AreEqual("localhost:143", ImapStyleUriParser.GetAuthority(new Uri("imap://localhost")));
      Assert.AreEqual("localhost:993", ImapStyleUriParser.GetAuthority(new Uri("imaps://localhost")));
      Assert.AreEqual("localhost:10143", ImapStyleUriParser.GetAuthority(new Uri("imap://localhost:10143")));
      Assert.AreEqual("localhost:10143", ImapStyleUriParser.GetAuthority(new Uri("imaps://localhost:10143")));
      Assert.AreEqual("user@localhost:10143", ImapStyleUriParser.GetAuthority(new Uri("imap://user@localhost:10143")));
      Assert.AreEqual("localhost:10143", ImapStyleUriParser.GetAuthority(new Uri("imap://;auth=*@localhost:10143")));
      Assert.AreEqual("user@localhost:10143", ImapStyleUriParser.GetAuthority(new Uri("imap://user;auth=*@localhost:10143")));
    }

    [Test]
    public void TestGetStrongAuthority()
    {
      Assert.AreEqual("localhost:143", ImapStyleUriParser.GetStrongAuthority(new Uri("imap://localhost")));
      Assert.AreEqual("user@localhost:143", ImapStyleUriParser.GetStrongAuthority(new Uri("imap://user@localhost")));
      Assert.AreEqual(";auth=*@localhost:143", ImapStyleUriParser.GetStrongAuthority(new Uri("imap://;auth=*@localhost")));
      Assert.AreEqual("user;auth=*@localhost:143", ImapStyleUriParser.GetStrongAuthority(new Uri("imap://user;auth=*@localhost")));
    }

    [Test]
    public void TestGetUser()
    {
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetUser(new Uri("imap://localhost:10143")));
      Assert.AreEqual("user", ImapStyleUriParser.GetUser(new Uri("imap://user@localhost:10143")));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetUser(new Uri("imap://;auth=*@localhost:10143")));
      Assert.AreEqual("user", ImapStyleUriParser.GetUser(new Uri("imap://user;auth=*@localhost:10143")));
    }

    [Test]
    public void TestGetAuthType()
    {
      Assert.AreEqual(null, ImapStyleUriParser.GetAuthType(new Uri("imap://localhost:10143")));
      Assert.AreEqual(null, ImapStyleUriParser.GetAuthType(new Uri("imap://user@localhost:10143")));
      Assert.AreEqual(ImapAuthenticationMechanism.SelectAppropriate, ImapStyleUriParser.GetAuthType(new Uri("imap://;auth=*@localhost:10143")));
      Assert.AreEqual(ImapAuthenticationMechanism.SelectAppropriate, ImapStyleUriParser.GetAuthType(new Uri("imap://;AUTH=*@localhost:10143")));
      Assert.AreEqual(ImapAuthenticationMechanism.SelectAppropriate, ImapStyleUriParser.GetAuthType(new Uri("imap://user;auth=*@localhost:10143")));
    }

    [Test]
    public void TestGetMailbox()
    {
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/")));
      Assert.AreEqual("INBOX", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX")));
      Assert.AreEqual("INBOX", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX/")));
      Assert.AreEqual("INBOX", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX;UIDVALIDITY=1")));
      Assert.AreEqual("INBOX", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX;UIDVALIDITY=1/")));
      Assert.AreEqual("INBOX/Sent", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX/Sent")));
      Assert.AreEqual("INBOX/Sent", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX/Sent;UIDVALIDITY=1")));
      Assert.AreEqual("INBOX.Sent", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX.Sent")));
      Assert.AreEqual("INBOX.Sent", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/INBOX.Sent;UIDVALIDITY=1")));
      Assert.AreEqual("新しいメールボックス", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/新しいメールボックス")));
      Assert.AreEqual("&ZbAwVzBEMOEw,DDrMNwwwzCvMLk-", ImapStyleUriParser.GetMailbox(new Uri("imap://localhost/&ZbAwVzBEMOEw,DDrMNwwwzCvMLk-")));
    }

    [Test]
    public void TestGetDecodedMailbox()
    {
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/")));
      Assert.AreEqual("新しいメールボックス", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/新しいメールボックス")));
      Assert.AreEqual("新しいメールボックス", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/新しいメールボックス;UIDVALIDITY=1")));
      Assert.AreEqual("新しいメールボックス", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/&ZbAwVzBEMOEw,DDrMNwwwzCvMLk-")));
      Assert.AreEqual("新しいメールボックス", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/&ZbAwVzBEMOEw,DDrMNwwwzCvMLk-;UIDVALIDITY=1")));
      Assert.AreEqual("&INBOX&", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/&INBOX&")));
      Assert.AreEqual("&INBOX&", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/&INBOX&;UIDVALIDITY=1")));
      Assert.AreEqual("&INBOX&", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/&-INBOX&-")));
      Assert.AreEqual("&INBOX&", ImapStyleUriParser.GetDecodedMailbox(new Uri("imap://localhost/&-INBOX&-;UIDVALIDITY=1")));
    }

    [Test]
    public void TestUidValidity()
    {
      Assert.AreEqual(0L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/")));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX;UIDVALIDITY=1")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX;uidvalidity=1")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX;UIDVALIDITY=1/")));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX/Sent")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX/Sent;UIDVALIDITY=1")));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX.Sent")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUidValidity(new Uri("imap://localhost/INBOX.Sent;UIDVALIDITY=1")));
    }

    [Test]
    public void TestGetUid()
    {
      Assert.AreEqual(0L, ImapStyleUriParser.GetUid(new Uri("imap://localhost/INBOX/")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUid(new Uri("imap://localhost/INBOX/;UID=1")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUid(new Uri("imap://localhost/INBOX/;uid=1")));
      Assert.AreEqual(1L, ImapStyleUriParser.GetUid(new Uri("imap://localhost/INBOX/;UID=1/;PARTIAL=0.1024")));
    }

    [Test]
    public void TestGetSection()
    {
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetSection(new Uri("imap://localhost/INBOX/;UID=1")));
      Assert.AreEqual("1.1",        ImapStyleUriParser.GetSection(new Uri("imap://localhost/INBOX/;UID=1/;SECTION=1.1")));
      Assert.AreEqual("1.1",        ImapStyleUriParser.GetSection(new Uri("imap://localhost/INBOX/;UID=1/;section=1.1")));
      Assert.AreEqual("1.1",        ImapStyleUriParser.GetSection(new Uri("imap://localhost/INBOX/;UID=1/;SECTION=1.1/")));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetSection(new Uri("imap://localhost/INBOX/;UID=1/;PARTIAL=0.1024")));
      Assert.AreEqual("1.1",        ImapStyleUriParser.GetSection(new Uri("imap://localhost/INBOX/;UID=1/;SECTION=1.1/;PARTIAL=0.1024")));
    }

    [Test]
    public void TestGetPartial()
    {
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetPartial(new Uri("imap://localhost/INBOX/;UID=1")));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetPartial(new Uri("imap://localhost/INBOX/;UID=1/;SECTION=1.1")));
      Assert.AreEqual("0.1024",     ImapStyleUriParser.GetPartial(new Uri("imap://localhost/INBOX/;UID=1/;PARTIAL=0.1024")));
      Assert.AreEqual("0.1024",     ImapStyleUriParser.GetPartial(new Uri("imap://localhost/INBOX/;UID=1/;partial=0.1024")));
      Assert.AreEqual("0.1024",     ImapStyleUriParser.GetPartial(new Uri("imap://localhost/INBOX/;UID=1/;PARTIAL=0.1024/")));
      Assert.AreEqual("0.1024",     ImapStyleUriParser.GetPartial(new Uri("imap://localhost/INBOX/;UID=1/;SECTION=1.1/;PARTIAL=0.1024")));
    }

    [Test]
    public void TestGetUriForm()
    {
      TestGetUriForm(ImapUriForm.Server, "imap://localhost/");
      TestGetUriForm(ImapUriForm.Server, "imap://localhost:10143/");
      TestGetUriForm(ImapUriForm.Server, "imap://user@localhost:10143/");
      TestGetUriForm(ImapUriForm.Server, "imap://;auth=*@localhost:10143/");
      TestGetUriForm(ImapUriForm.Server, "imap://user;auth=*@localhost:10143/");

      TestGetUriForm(ImapUriForm.ListMessages, "imap://localhost/INBOX/");
      TestGetUriForm(ImapUriForm.ListMessages, "imap://localhost/INBOX/Sent/");
      TestGetUriForm(ImapUriForm.ListMessages, "imap://localhost/INBOX.Sent/");
      TestGetUriForm(ImapUriForm.ListMessages, "imap://localhost/INBOX/Sent;uidvalidity=1/");
      TestGetUriForm(ImapUriForm.ListMessages, "imap://localhost/INBOX.Sent;uidvalidity=1/");

      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX/;uid=1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX/;uid=1/;section=1.1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX/;uid=1/;partial=1.1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX/;uid=1/;section=1.1/;partial=1.1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX;uidvalidity=1/;uid=1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX/Sent/;uid=1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX.Sent/;uid=1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX/Sent;uidvalidity=1/;uid=1");
      TestGetUriForm(ImapUriForm.FetchMessage, "imap://localhost/INBOX.Sent;uidvalidity=1/;uid=1");

      TestGetUriForm(ImapUriForm.SearchMessages, "imap://localhost/INBOX/?UID 1");
      TestGetUriForm(ImapUriForm.SearchMessages, "imap://localhost/INBOX;uidvalidity=1/?UID 1");
    }

    private void TestGetUriForm(ImapUriForm expected, string uri)
    {
      Assert.AreEqual(expected, ImapStyleUriParser.GetUriForm(new Uri(uri)), uri);
    }

    [Test()]
    public void TestImapUrlExamples1()
    {
      //   imap://<iserver>/
      var uri = new Uri("imap://localhost:993/");

      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetUser(uri));
      Assert.AreEqual("localhost:993", ImapStyleUriParser.GetAuthority(uri));
      Assert.AreEqual("localhost:993", ImapStyleUriParser.GetStrongAuthority(uri));
      Assert.AreEqual(null, ImapStyleUriParser.GetAuthType(uri));
      Assert.AreEqual(ImapUriForm.Server, ImapStyleUriParser.GetUriForm(uri));

      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetMailbox(uri));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUidValidity(uri));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUid(uri));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetSection(uri));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetPartial(uri));
    }

    [Test()]
    public void TestImapUrlExamples2()
    {
      var uri = new Uri("imap://user;auth=plain@localhost");

      Assert.AreEqual("user", ImapStyleUriParser.GetUser(uri));
      Assert.AreEqual("user@localhost:143", ImapStyleUriParser.GetAuthority(uri));
      Assert.AreEqual("user;auth=plain@localhost:143", ImapStyleUriParser.GetStrongAuthority(uri));
      Assert.AreEqual(ImapAuthenticationMechanism.Plain, ImapStyleUriParser.GetAuthType(uri));
      Assert.AreEqual(ImapUriForm.Server, ImapStyleUriParser.GetUriForm(uri));

      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetMailbox(uri));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUidValidity(uri));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUid(uri));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetSection(uri));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetPartial(uri));
    }

    [Test()]
    public void TestImapUrlExamples3()
    {
      //   imap://<iserver>/<enc_mailbox>[uidvalidity]<iuid>[isection]
      var uri = new Uri("imap://psicorp.org/~peter/%E6%97%A5%E6%9C%AC%E8%AA%9E/%E5%8F%B0%E5%8C%97");

      Assert.AreEqual("psicorp.org:143", ImapStyleUriParser.GetAuthority(uri));
      Assert.AreEqual("~peter/日本語/台北", ImapStyleUriParser.GetMailbox(uri));
      Assert.AreEqual(ImapUriForm.ListMessages, ImapStyleUriParser.GetUriForm(uri));
    }

    [Test()]
    public void TestImapUrlExamples4()
    {
      var uri = new Uri("imap://minbari.example.org/gray-council;UIDVALIDITY=385759045/;UID=20/;PARTIAL=0.1024");

      Assert.AreEqual("minbari.example.org:143", ImapStyleUriParser.GetAuthority(uri));
      Assert.AreEqual("gray-council", ImapStyleUriParser.GetMailbox(uri));
      Assert.AreEqual(385759045L, ImapStyleUriParser.GetUidValidity(uri));
      Assert.AreEqual(20L, ImapStyleUriParser.GetUid(uri));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetSection(uri));
      Assert.AreEqual("0.1024", ImapStyleUriParser.GetPartial(uri));
      Assert.AreEqual(ImapUriForm.FetchMessage, ImapStyleUriParser.GetUriForm(uri));
    }

    [Test()]
    public void TestImapUrlExamples5()
    {
      var uri = new Uri("imap://;AUTH=KERBEROS_V4@minbari.org/gray-council/;uid=20/;section=1.2");

      Assert.AreEqual("minbari.org:143", ImapStyleUriParser.GetAuthority(uri));
      Assert.AreEqual(";AUTH=KERBEROS_V4@minbari.org:143", ImapStyleUriParser.GetStrongAuthority(uri));
      Assert.AreEqual(ImapAuthenticationMechanism.KerberosV4, ImapStyleUriParser.GetAuthType(uri));
      Assert.AreEqual("gray-council", ImapStyleUriParser.GetMailbox(uri));
      Assert.AreEqual(0L, ImapStyleUriParser.GetUidValidity(uri));
      Assert.AreEqual(20L, ImapStyleUriParser.GetUid(uri));
      Assert.AreEqual("1.2", ImapStyleUriParser.GetSection(uri));
      Assert.AreEqual(string.Empty, ImapStyleUriParser.GetPartial(uri));
      Assert.AreEqual(ImapUriForm.FetchMessage, ImapStyleUriParser.GetUriForm(uri));
    }
  }
}
