// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Net;

namespace Smdn.Interop.Sspi {
  [CLSCompliant(false)]
  public abstract class SspiBase : IDisposable {
    ~SspiBase()
    {
      Dispose(false);
    }

    public void Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    protected virtual void Dispose(bool disposing)
    {
      if (hasCredential) {
        secur32.FreeCredentialsHandle(ref hCred);
        hasCredential = false;
      }

      disposed = true;
    }

    public void AcquireCredentials(NetworkCredential credential, string securityPackageName)
    {
      CheckDisposed();

      if (securityPackageName == null)
        throw new ArgumentNullException("securityPackageName");

      var fCredentialUse = (this is SspiClient) ? SECPKG_CRED.SECPKG_CRED_OUTBOUND : SECPKG_CRED.SECPKG_CRED_INBOUND;

      if (credential == null) {
        SspiException.ThrowIfError(secur32.AcquireCredentialsHandle(System.Security.Principal.WindowsIdentity.GetCurrent().Name,
                                                                    securityPackageName,
                                                                    fCredentialUse,
                                                                    IntPtr.Zero,
                                                                    IntPtr.Zero,
                                                                    0,
                                                                    IntPtr.Zero,
                                                                    out hCred,
                                                                    out tsExpiry));
      }
      else {
        var pAuthData = new SEC_WINNT_AUTH_IDENTITY();

        pAuthData.User            = credential.UserName;
        pAuthData.UserLength      = (uint)(credential.UserName ?? string.Empty).Length;
        pAuthData.Password        = credential.Password;
        pAuthData.PasswordLength  = (uint)(credential.Password ?? string.Empty).Length;
        pAuthData.Domain          = credential.Domain;
        pAuthData.DomainLength    = (uint)(credential.Domain ?? string.Empty).Length;
        pAuthData.Flags           = SEC_WINNT_AUTH_IDENTITY.SEC_WINNT_AUTH_IDENTITY_UNICODE;

        SspiException.ThrowIfError(secur32.AcquireCredentialsHandle(null,
                                                                    securityPackageName,
                                                                    fCredentialUse,
                                                                    IntPtr.Zero,
                                                                    ref pAuthData,
                                                                    0,
                                                                    IntPtr.Zero,
                                                                    out hCred,
                                                                    out tsExpiry));
      }

      hasCredential = true;
    }

    public bool InitializeSecurityContext(byte[] inputToken, out byte[] outputToken)
    {
      CheckDisposed();

      if (!hasCredential)
        throw new InvalidOperationException("call AcquireCredentials first");

      return (InitializeSecurityContext(hCred, inputToken, out outputToken) == SECURITY_STATUS.SEC_E_OK);
    }

    protected abstract SECURITY_STATUS InitializeSecurityContext(CredHandle hCred, byte[] inputToken, out byte[] outputToken);

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private bool disposed = false;
    private CredHandle hCred;
    private bool hasCredential = false;
    protected TimeStamp tsExpiry;
  }
}
