// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#define MONO_SECURITY_DLL
#undef MONO_SECURITY_DLL

using System;
using System.Net;

using Smdn.Interop.Sspi;

#if MONO_SECURITY_DLL
using Mono.Security.Protocol.Ntlm;
#else
using System.Reflection;
#endif

namespace Smdn.Security.Authentication.Sasl.Client {
  internal abstract class NTLMMechanismImpl : IDisposable {
    public NetworkCredential Credential {
      get; set;
    }

    public string TargetHost {
      get; set;
    }

    public abstract void Initialize();
    public abstract void Dispose();
    public abstract void Negotiate(byte[] serverChallenge, out byte[] clientResponse);
    public abstract void Authenticate(byte[] serverChallenge, out byte[] clientResponse);

    public static NTLMMechanismImpl Create()
    {
      try {
        return new SspiNTLMMechanismImpl();
      }
      catch (NotSupportedException) {
        return new MonoNTLMMechanismImpl();
      }
    }
  }

  internal class MonoNTLMMechanismImpl : NTLMMechanismImpl {
    public MonoNTLMMechanismImpl()
    {
#if !MONO_SECURITY_DLL
      try {
        var assm = Assembly.Load("Mono.Security, Version=2.0.0.0, Culture=neutral, PublicKeyToken=0738eb9f132ed756");

        typeOfType1Message = assm.GetType("Mono.Security.Protocol.Ntlm.Type1Message", true);
        typeOfType2Message = assm.GetType("Mono.Security.Protocol.Ntlm.Type2Message", true);
        typeOfType3Message = assm.GetType("Mono.Security.Protocol.Ntlm.Type3Message", true);
      }
      catch (System.IO.FileNotFoundException) {
        throw new PlatformNotSupportedException();
      }
      catch (System.IO.FileLoadException) {
        throw new PlatformNotSupportedException();
      }
      catch (BadImageFormatException) {
        throw new PlatformNotSupportedException();
      }
      catch (TypeLoadException) {
        throw new PlatformNotSupportedException();
      }
#endif
    }

    public override void Initialize()
    {
      // nothing to do
    }

    public override void Dispose()
    {
      // nothing to do
    }

    public override void Negotiate(byte[] serverChallenge, out byte[] clientResponse)
    {
#if MONO_SECURITY_DLL
      var type1 = new Type1Message();

      type1.Host = TargetHost ?? string.Empty; // ?
      type1.Domain = Credential.Domain ?? string.Empty;

      clientResponse = type1.GetBytes();
#else
      var type1 = Activator.CreateInstance(typeOfType1Message);

      typeOfType1Message.GetProperty("Host").SetValue(type1, TargetHost ?? string.Empty, null); // ?
      typeOfType1Message.GetProperty("Domain").SetValue(type1, Credential.Domain ?? string.Empty, null);

      clientResponse = (byte[])typeOfType1Message.GetMethod("GetBytes").Invoke(type1, null);
#endif
    }

    public override void Authenticate(byte[] serverChallenge, out byte[] clientResponse)
    {
#if MONO_SECURITY_DLL
      var type2 = new Type2Message(serverChallenge);
      var type3 = new Type3Message();

      type3.Host = TargetHost ?? string.Empty; // ?
      type3.Domain = Credential.Domain ?? string.Empty;

      type3.Challenge = type2.Nonce;
      type3.Password = Credential.Password ?? string.Empty;
      type3.Username = Credential.UserName;

      clientResponse = type3.GetBytes();
#else
      var type2 = Activator.CreateInstance(typeOfType2Message, serverChallenge);
      var type3 = Activator.CreateInstance(typeOfType3Message);

      var nonce = typeOfType2Message.GetProperty("Nonce").GetValue(type2, null) as byte[];

      typeOfType3Message.GetProperty("Host").SetValue(type3, TargetHost ?? string.Empty, null); // ?
      typeOfType3Message.GetProperty("Domain").SetValue(type3, Credential.Domain ?? string.Empty, null);

      typeOfType3Message.GetProperty("Challenge").SetValue(type3, nonce, null);
      typeOfType3Message.GetProperty("Password").SetValue(type3, Credential.Password ?? string.Empty, null);
      typeOfType3Message.GetProperty("Username").SetValue(type3, Credential.UserName, null);

      clientResponse = (byte[])typeOfType3Message.GetMethod("GetBytes").Invoke(type3, null);
#endif
    }

#if !MONO_SECURITY_DLL
    private /*readonly*/ Type typeOfType1Message;
    private /*readonly*/ Type typeOfType2Message;
    private /*readonly*/ Type typeOfType3Message;
#endif
  }

  internal class SspiNTLMMechanismImpl : NTLMMechanismImpl {
    public SspiNTLMMechanismImpl()
    {
      try {
        SecPkgInfo.QuerySecurityPackageInfo("NTLM");
      }
      catch (TypeLoadException) { // DllNotFoundException, EntryPointNotFoundException
        throw new PlatformNotSupportedException();
      }
      catch (SspiException) {
        throw new NotSupportedException();
      }
    }

    public override void Initialize()
    {
      if (client != null)
        client.Dispose();

      client = new SspiClient(TargetHost);
    }

    public override void Dispose()
    {
      if (client != null) {
        client.Dispose();
        client = null;
      }
    }

    public override void Negotiate(byte[] serverChallenge, out byte[] clientResponse)
    {
      client.AcquireCredentials(base.Credential, "NTLM");
      client.InitializeSecurityContext(serverChallenge, out clientResponse);
    }

    public override void Authenticate(byte[] serverChallenge, out byte[] clientResponse)
    {
      client.InitializeSecurityContext(serverChallenge, out clientResponse);
    }

    private SspiClient client;
  }
}
